import inspect
import sys

from .base_backend import BaseBackend
from .lime_backend import LimeBackend
from .shap_backend import ShapBackend


def get_backend_cls_from_name(name):
    """
    Scan current module to find the right backend with given name.
    """
    list_cls = [
        cls
        for _, cls in inspect.getmembers(sys.modules[__name__])
        if (
            inspect.isclass(cls)
            and issubclass(cls, BaseBackend)
            and cls.name.lower() == name.lower()
            and cls.name.lower() != "base"
        )
    ]

    if len(list_cls) == 1:
        return list_cls[0]
    else:
        raise ValueError(f"Backend class not found with name : {name}")



import importlib

# This list should be identical to the list in setup.py
report_requirements = ["nbconvert==6.0.7", "papermill", "matplotlib", "seaborn", "notebook", "Jinja2"]


def check_report_requirements():
    """
    Checks that all required packages for the report are installed.
    This function should be called before executing the report.
    """
    for req in report_requirements:
        pkg = req.split("=")[0]
        try:
            importlib.import_module(pkg.lower())
        except ImportError:
            raise ModuleNotFoundError(
                f"The following package is necessary to generate the Shapash Report : {pkg}. "
                f"Try 'pip install shapash[report]' to install all required packages."
            )



"""
Model Synoptic Module
"""

simple_tree_model_sklearn = (
    "<class 'sklearn.ensemble._forest.ExtraTreesClassifier'>",
    "<class 'sklearn.ensemble._forest.ExtraTreesRegressor'>",
    "<class 'sklearn.ensemble._forest.RandomForestClassifier'>",
    "<class 'sklearn.ensemble._forest.RandomForestRegressor'>",
    "<class 'sklearn.ensemble._gb.GradientBoostingClassifier'>",
    "<class 'sklearn.ensemble._gb.GradientBoostingRegressor'>",
)
xgboost_model = (
    "<class 'xgboost.sklearn.XGBClassifier'>",
    "<class 'xgboost.sklearn.XGBRegressor'>",
    "<class 'xgboost.core.Booster'>",
)

lightgbm_model = (
    "<class 'lightgbm.sklearn.LGBMClassifier'>",
    "<class 'lightgbm.sklearn.LGBMRegressor'>",
    "<class 'lightgbm.basic.Booster'>",
)

catboost_model = ("<class 'catboost.core.CatBoostClassifier'>", "<class 'catboost.core.CatBoostRegressor'>")

linear_model = (
    "<class 'sklearn.linear_model._logistic.LogisticRegression'>",
    "<class 'sklearn.linear_model._base.LinearRegression'>",
)

svm_model = ("<class 'sklearn.svm._classes.SVC'>", "<class 'sklearn.svm._classes.SVR'>")

simple_tree_model = simple_tree_model_sklearn + xgboost_model + lightgbm_model

dict_model_feature = {
    "<class 'sklearn.ensemble._forest.ExtraTreesClassifier'>": ["length"],
    "<class 'sklearn.ensemble._forest.ExtraTreesRegressor'>": ["length"],
    "<class 'sklearn.ensemble._forest.RandomForestClassifier'>": ["length"],
    "<class 'sklearn.ensemble._forest.RandomForestRegressor'>": ["length"],
    "<class 'sklearn.ensemble._gb.GradientBoostingClassifier'>": ["length"],
    "<class 'sklearn.ensemble._gb.GradientBoostingRegressor'>": ["length"],
    "<class 'sklearn.linear_model._logistic.LogisticRegression'>": ["length"],
    "<class 'sklearn.linear_model._base.LinearRegression'>": ["length"],
    "<class 'sklearn.svm._classes.SVC'>": ["length"],
    "<class 'sklearn.svm._classes.SVR'>": ["length"],
    "<class 'lightgbm.sklearn.LGBMClassifier'>": ["booster_", "feature_name"],
    "<class 'lightgbm.sklearn.LGBMRegressor'>": ["booster_", "feature_name"],
    "<class 'lightgbm.basic.Booster'>": ["feature_names"],
    "<class 'xgboost.sklearn.XGBClassifier'>": ["get_booster", "feature_names"],
    "<class 'xgboost.sklearn.XGBRegressor'>": ["get_booster", "feature_names"],
    "<class 'xgboost.core.Booster'>": ["feature_names"],
    "<class 'catboost.core.CatBoostClassifier'>": ["feature_names_"],
    "<class 'catboost.core.CatBoostRegressor'>": ["feature_names_"],
}



"""Top-level package."""

__author__ = (
    """Yann Golhen, Yann Lagré, Sebastien Bidault, Maxime Gendre, Thomas Bouche, Johann Martin, Guillaume Vignal"""
)
__email__ = "yann.golhen@maif.fr, yann.lagre@maif.fr, sebabstien.bidault.marketing@maif.fr, thomas.bouche@maif.fr, guillaume.vignal@maif.fr"

from shapash.explainer.smart_explainer import SmartExplainer

from .__version__ import __version__



"""
Main class of Web application Shapash
"""
import copy
import random
import re
from math import log10

import dash
import dash_bootstrap_components as dbc
import dash_daq as daq
import pandas as pd
import plotly.graph_objs as go
from dash import ALL, MATCH, dash_table, dcc, html
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from flask import Flask

from shapash.utils.utils import truncate_str
from shapash.webapp.utils.callbacks import (
    create_dropdown_feature_filter,
    create_filter_modalities_selection,
    create_id_card_data,
    create_id_card_layout,
    get_feature_contributions_sign_to_show,
    get_feature_filter_options,
    get_feature_from_clicked_data,
    get_feature_from_features_groups,
    get_figure_zoom,
    get_group_name,
    get_id_card_contrib,
    get_id_card_features,
    get_indexes_from_datatable,
    select_data_from_bool_filters,
    select_data_from_date_filters,
    select_data_from_numeric_filters,
    select_data_from_prediction_picking,
    select_data_from_str_filters,
    update_click_data_on_subset_changes,
    update_features_to_display,
)
from shapash.webapp.utils.explanations import Explanations
from shapash.webapp.utils.MyGraph import MyGraph
from shapash.webapp.utils.utils import check_row, get_index_type, round_to_k


def _create_input_modal(id, label, tooltip):
    return dbc.Row(
        [
            dbc.Label(label, id=f"{id}_label", html_for=id, width=8),
            dbc.Col(dbc.Input(id=id, type="number", value=0), width=4),
            dbc.Tooltip(tooltip, target=f"{id}_label", placement="bottom"),
        ],
        className="g-3",
    )


class SmartApp:
    """
    Bridge pattern decoupling the application part from SmartExplainer and SmartPlotter.
    Attributes
    ----------
    explainer: object
        SmartExplainer instance to point to.
    """

    def __init__(self, explainer, settings: dict = None):
        """
        Init on class instantiation, everything to be able to run the app on server.
        Parameters
        ----------
        explainer : SmartExplainer
            SmartExplainer object
        settings : dict
            A dict describing the default webapp settings values to be used
            Possible settings (dict keys) are 'rows', 'points', 'violin', 'features'
            Values should be positive ints
        """
        # APP
        self.server = Flask(__name__)
        self.app = dash.Dash(
            server=self.server,
            external_stylesheets=[dbc.themes.BOOTSTRAP],
        )
        self.app.title = "Shapash Monitor"
        if explainer.title_story:
            self.app.title += " - " + explainer.title_story
        self.explainer = explainer

        # SETTINGS
        self.logo = self.app.get_asset_url("shapash-fond-fonce.png")
        self.color = explainer.plot._style_dict["webapp_button"]
        self.bkg_color = explainer.plot._style_dict["webapp_bkg"]
        self.title_menu_color = explainer.plot._style_dict["webapp_title"]
        self.settings_ini = {
            "rows": 1000,
            "points": 1000,
            "violin": 10,
            "features": 20,
        }
        if settings is not None:
            for k, v in self.settings_ini.items():
                self.settings_ini[k] = (
                    settings[k] if k in settings and isinstance(settings[k], int) and 0 < settings[k] else v
                )
        self.settings = self.settings_ini.copy()

        self.predict_col = ["_predict_"]
        self.special_cols = ["_index_", "_predict_"]
        if self.explainer.y_target is not None:
            self.special_cols.append("_target_")
            if self.explainer._case == "regression":
                self.special_cols.append("_error_")
        self.explainer.features_imp = self.explainer.state.compute_features_import(self.explainer.contributions)
        if self.explainer._case == "classification":
            self.label = self.explainer.check_label_name(len(self.explainer._classes) - 1, "num")[1]
            self.selected_feature = self.explainer.features_imp[-1].idxmax()
            self.max_threshold = int(
                max([x.map(lambda x: round_to_k(x, k=1)).max().max() for x in self.explainer.contributions])
            )
        else:
            self.label = None
            self.selected_feature = self.explainer.features_imp.idxmax()
            self.max_threshold = int(self.explainer.contributions.map(lambda x: round_to_k(x, k=1)).max().max())
        self.list_index = []
        self.subset = None
        self.last_click_data = None

        # DATA
        self.explanations = Explanations()  # To get explanations of "?" buttons
        self.dataframe = pd.DataFrame()
        self.round_dataframe = pd.DataFrame()
        self.features_dict = copy.deepcopy(self.explainer.features_dict)
        self.init_data()

        # COMPONENTS
        self.components = {"menu": {}, "table": {}, "graph": {}, "filter": {}, "settings": {}}
        self.init_components()

        # LAYOUT
        self.skeleton = {"navbar": {}, "body": {}}
        self.make_skeleton()
        self.app.layout = html.Div([self.skeleton["navbar"], self.skeleton["body"]])

        # CALLBACK
        self.callback_fullscreen_buttons()
        self.init_callback_settings()
        self.callback_generator()

    def init_data(self, rows=None):
        """
        Method which initializes data from explainer object
        """
        if hasattr(self.explainer, "y_pred"):
            self.dataframe = self.explainer.x_init.copy()
            if isinstance(self.explainer.y_pred, (pd.Series, pd.DataFrame)):
                self.predict_col = self.explainer.y_pred.columns.to_list()[0]
                self.dataframe = self.dataframe.join(self.explainer.y_pred)
            elif isinstance(self.explainer.y_pred, list):
                self.dataframe = self.dataframe.join(
                    pd.DataFrame(
                        data=self.explainer.y_pred, columns=[self.predict_col], index=self.explainer.x_init.index
                    )
                )
            else:
                raise TypeError("y_pred must be of type pd.Series, pd.DataFrame or list")
        else:
            raise ValueError("y_pred must be set when calling compile function.")

        if self.explainer.additional_data is not None:
            self.dataframe = self.dataframe.join(self.explainer.additional_data)
            self.features_dict.update(self.explainer.additional_features_dict)

        self.dataframe["_index_"] = self.explainer.x_init.index
        self.dataframe.rename(columns={f"{self.predict_col}": "_predict_"}, inplace=True)
        if self.explainer.y_target is not None:
            self.dataframe = self.dataframe.join(
                self.explainer.y_target.rename(columns={self.explainer.y_target.columns[0]: "_target_"}),
            )
            if self.explainer._case == "regression":
                self.dataframe = self.dataframe.join(self.explainer.prediction_error)

        col_order = self.special_cols + self.dataframe.columns.drop(self.special_cols).tolist()
        random.seed(79)
        if rows is None:
            rows = self.settings["rows"]
        self.list_index = random.sample(
            population=self.dataframe.index.tolist(), k=min(rows, len(self.dataframe.index.tolist()))
        )
        self.dataframe = self.dataframe[col_order].loc[self.list_index].sort_index()
        self.round_dataframe = self.dataframe.copy()
        for col in list(self.dataframe.columns):
            typ = self.dataframe[col].dtype
            if typ == float:
                std = self.dataframe[col].std()
                if std != 0:
                    digit = max(round(log10(1 / std) + 1) + 2, 0)
                    self.round_dataframe[col] = self.dataframe[col].map(f"{{:.{digit}f}}".format).astype(float)

    def init_components(self):
        """
        Initialize components (graph, table, filter, settings, ...) and insert it inside
        components containers which are created by init_skeleton
        """

        self.components["settings"]["input_rows"] = _create_input_modal(
            id="rows",
            label="Number of rows for subset",
            tooltip="Set max number of lines for subset (datatable). \
                     Filter will be apply on this subset.",
        )

        self.components["settings"]["input_points"] = _create_input_modal(
            id="points",
            label="Number of points for plot",
            tooltip="Set max number of points in feature contribution plots.",
        )

        self.components["settings"]["input_features"] = _create_input_modal(
            id="features",
            label="Number of features to plot",
            tooltip="Set max number of features to plot in features \
                     importance and local explanation plots.",
        )

        self.components["settings"]["input_violin"] = _create_input_modal(
            id="violin",
            label="Max number of labels for violin plot",
            tooltip="Set max number of labels to display a violin plot \
                     for feature contribution plot (otherwise a scatter \
                                                    plot is displayed).",
        )

        self.components["settings"]["name"] = dbc.Row(
            [
                dbc.Checklist(
                    options=[
                        {
                            "label": "Use domain name for \
                              features name.",
                            "value": 1,
                        }
                    ],
                    value=[],
                    inline=True,
                    id="name",
                    style={"margin-left": "20px"},
                ),
                dbc.Tooltip(
                    "Replace technical feature names by \
                            domain names if exists.",
                    target="name",
                    placement="bottom",
                ),
            ],
            className="g-3",
        )

        self.components["settings"]["modal"] = dbc.Modal(
            [
                dbc.ModalHeader("Settings"),
                dbc.ModalBody(
                    dbc.Form(
                        [
                            self.components["settings"]["input_rows"],
                            self.components["settings"]["input_points"],
                            self.components["settings"]["input_features"],
                            self.components["settings"]["input_violin"],
                            self.components["settings"]["name"],
                        ]
                    )
                ),
                dbc.ModalFooter(dbc.Button("Apply", id="apply", className="ml-auto")),
            ],
            id="modal",
        )

        self.components["menu"] = dbc.Row(
            [
                dbc.Col(
                    [
                        html.Div(
                            daq.BooleanSwitch(
                                id="bool_groups",
                                on=True,
                                style={"display": "none"} if self.explainer.features_groups is None else {},
                                color=self.color[0],
                                label={
                                    "label": "Groups",
                                    "style": {
                                        "fontSize": 18,
                                        "color": self.color[0],
                                        "fontWeight": "bold",
                                        "margin-left": "5px",
                                    },
                                },
                                labelPosition="right",
                            ),
                            style={"margin-right": "35px"},
                        )
                    ],
                    width="auto",
                    align="center",
                ),
                dbc.Col(
                    [
                        html.H4(
                            [
                                dbc.Badge(
                                    "Regression",
                                    id="regression_badge",
                                    style={"margin-right": "5px", "margin-left": "0px"},
                                    color="",
                                ),
                                dbc.Badge("Classification", id="classification_badge", color=""),
                            ],
                            style={"margin-right": "5px"},
                        ),
                    ],
                    width="auto",
                    align="center",
                    style={"padding": "auto"},
                ),
                dbc.Col(
                    dbc.Collapse(
                        dbc.Row(
                            [
                                # 2 columns to have class beside the dropdown buttons
                                dbc.Col(
                                    [
                                        dbc.Label("Class:", style={"color": "white", "margin": "0px"}),
                                    ],
                                    align="center",
                                ),
                                dbc.Col(
                                    [
                                        dcc.Dropdown(
                                            id="select_label",
                                            options=[],
                                            value=None,
                                            clearable=False,
                                            searchable=False,
                                            style={
                                                "verticalAlign": "middle",
                                                "zIndex": "1010",
                                                "min-width": "160px",
                                                "height": "100%",
                                            },
                                        )
                                    ],
                                    style={"margin-right": "17px", "padding": "0px", "width": "auto"},
                                ),
                            ],
                            style={"margin": "0px"},
                        ),
                        is_open=True,
                        id="select_collapse",
                    ),
                    width="auto",
                    align="center",
                    style={"padding": "0px"},
                ),
                dbc.Col(
                    [
                        html.Div(
                            [
                                html.Img(
                                    id="settings",
                                    title="settings",
                                    alt="Settings",
                                    src=self.app.get_asset_url("settings.png"),
                                    height="40px",
                                    style={"cursor": "pointer"},
                                ),
                                self.components["settings"]["modal"],
                            ]
                        )
                    ],
                    align="center",
                    width="auto",
                    style={"padding": "0px"},
                ),
            ],
            className="g-0",
            justify="end",
        )

        self.adjust_menu()

        self.components["table"]["dataset"] = dash_table.DataTable(
            id="dataset",
            data=self.round_dataframe.to_dict("records"),
            tooltip_data=[
                {column: {"value": str(value), "type": "text"} for column, value in row.items()}
                for row in self.dataframe.to_dict("index").values()
            ],
            tooltip_duration=2000,
            columns=[{"name": i, "id": i} for i in self.dataframe.columns],
            tooltip_header={
                column: self.features_dict[column]
                for column in self.dataframe.columns
                if column not in self.special_cols
            },
            editable=False,
            row_deletable=False,
            virtualization=True,
            page_action="none",
            fixed_rows={"headers": True, "data": 0},
            fixed_columns={"headers": True, "data": 0},
            sort_action="custom",
            sort_mode="multi",
            sort_by=[],
            style_table={"overflowY": "auto", "overflowX": "auto"},
            style_header={"height": "30px"},
            style_cell={
                "minWidth": "70px",
                "width": "120px",
                "maxWidth": "200px",
                "textOverflow": "ellipsis",
            },
        )

        self.components["graph"]["global_feature_importance"] = MyGraph(
            figure=go.Figure(), id="global_feature_importance"
        )

        self.components["graph"]["feature_selector"] = MyGraph(figure=go.Figure(), id="feature_selector")

        # Component for the graph prediction picking
        self.components["graph"]["prediction_picking"] = MyGraph(figure=go.Figure(), id="prediction_picking")

        self.components["graph"]["detail_feature"] = MyGraph(figure=go.Figure(), id="detail_feature")

        # Component create to filter the dataset
        self.components["filter"]["filter_dataset"] = dbc.Col(
            [
                dbc.Row(
                    html.Div(
                        id="main",
                        children=[
                            html.Div(
                                id="filters",
                                children=[
                                    # Create Add Filter button
                                    dbc.Button(
                                        id="add_dropdown_button",
                                        children="Add Filter",
                                        color="warning",
                                        size="sm",
                                        style={"margin-right": "20px"},
                                    ),
                                    # Create reset Filter button (disabled of no filters applied)
                                    dbc.Button(
                                        id="reset_dropdown_button",
                                        children="Reset all existing filters",
                                        color="warning",
                                        disabled=True,
                                        size="sm",
                                        style={"margin-right": "20px"},
                                    ),
                                    # Create explanation button
                                    dbc.Button(
                                        "?",
                                        id="open_filter",
                                        size="sm",
                                        color="warning",
                                    ),
                                    # Create popover on the explanation button
                                    dbc.Popover(
                                        "Click here to know how you can apply filters.",
                                        target="open_filter",
                                        body=True,
                                        trigger="hover",
                                    ),
                                    # Modal associated to the explanation button
                                    dbc.Modal(
                                        [
                                            dbc.ModalHeader(dbc.ModalTitle("Filters explanation")),
                                            dbc.ModalBody([html.Div(dcc.Markdown(self.explanations.filter))]),
                                            dbc.ModalFooter(dbc.Button("Close", id="close_filter", color="warning")),
                                        ],
                                        id="modal_filter",
                                        centered=True,
                                        size="lg",
                                    ),
                                    # Div which will contains the filters
                                    html.Div(id="dropdowns_container", children=[]),
                                ],
                            )
                        ],
                    )
                ),
                dbc.Row(
                    html.Div(
                        [
                            html.Br(),
                            # Create Apply Filter button (Hidden if no filter to apply)
                            dbc.Button(
                                id="apply_filter",
                                children="Apply filters",
                                color="warning",
                                size="sm",
                                style={"display": "none"},
                            ),
                            dbc.Label(
                                id="filtered_subset_info", children="Test", width="auto", style={"display": "none"}
                            ),
                        ],
                    )
                ),
            ],
            style={"maxheight": "22rem", "height": "21rem", "zIndex": 800},
        )

        self.components["filter"]["index"] = dbc.Col(
            dbc.Row(
                [
                    dbc.Col(
                        [
                            dbc.Input(
                                id="index_id",
                                type=get_index_type(self.dataframe),
                                size="s",
                                placeholder="_index_",
                                debounce=True,
                                persistence=True,
                                style={"textAlign": "right"},
                            )
                        ],
                        width={"size": 5},
                        style={"padding": "0px"},
                    ),
                    dbc.Col(
                        [
                            html.Img(
                                id="validation",
                                alt="Validate",
                                title="Validate index",
                                src=self.app.get_asset_url("reload.png"),
                                height="30px",
                                style={"cursor": "pointer"},
                            )
                        ],
                        width={"size": 1},
                        style={"padding": "0px"},
                        align="center",
                    ),
                    dbc.Col(
                        [
                            dbc.Button(
                                "ID Card",
                                id="id_card",
                                color="warning",
                                style={"display": "none"},
                            ),
                            dbc.Popover(
                                "Click here to visualize the identity card of the selected sample.",
                                target="id_card",
                                body=True,
                                trigger="hover",
                            ),
                            dbc.Modal(
                                [
                                    dbc.ModalHeader(
                                        dbc.Col(
                                            [
                                                dbc.ModalTitle("Identity Card"),
                                                dbc.Row(
                                                    [
                                                        dbc.Label("Sort by:", align="center", width="auto"),
                                                        dbc.Col(
                                                            [
                                                                dcc.Dropdown(
                                                                    id="select_id_card_sorting",
                                                                    options=[
                                                                        {"label": "Label", "value": "feature_name"},
                                                                        {
                                                                            "label": "Contribution",
                                                                            "value": "feature_contrib",
                                                                        },
                                                                    ],
                                                                    value="feature_name",
                                                                    clearable=False,
                                                                    searchable=False,
                                                                ),
                                                            ],
                                                            width=3,
                                                        ),
                                                        dbc.Label("Order:", align="center", width="auto"),
                                                        dbc.Col(
                                                            [
                                                                dcc.Dropdown(
                                                                    id="select_id_card_order",
                                                                    options=[
                                                                        {"label": "Ascending", "value": True},
                                                                        {"label": "Descending", "value": False},
                                                                    ],
                                                                    value=True,
                                                                    clearable=False,
                                                                    searchable=False,
                                                                ),
                                                            ],
                                                            width=3,
                                                        ),
                                                    ],
                                                    style={"margin-top": "0.5rem"},
                                                ),
                                                dbc.Row(
                                                    [
                                                        dbc.Label("Label", width=3, style={"fontWeight": "bold"}),
                                                        dbc.Label("Value", width=5, style={"fontWeight": "bold"}),
                                                        dbc.Col(width=1),
                                                        dbc.Label(
                                                            "Contribution",
                                                            id="id_card_title_contrib",
                                                            width=3,
                                                            style={"fontWeight": "bold"},
                                                        ),
                                                    ],
                                                    style={"margin-top": "0.5rem", "margin-bottom": "-1rem"},
                                                ),
                                            ]
                                        ),
                                        close_button=False,
                                    ),
                                    dbc.ModalBody(id="id_card_body"),
                                    dbc.ModalFooter(dbc.Button("Close", id="close_id_card", color="warning")),
                                ],
                                id="modal_id_card",
                                centered=True,
                                size="xl",
                                scrollable=True,
                            ),
                        ],
                        width=5,
                    ),
                ],
                justify="center",
            )
        )

        self.components["filter"]["threshold"] = dbc.Col(
            [
                dbc.Label("Threshold", html_for="slider", id="threshold_label"),
                dcc.Slider(
                    min=0,
                    max=self.max_threshold,
                    value=0,
                    step=0.1,
                    marks={
                        f"{round(self.max_threshold * mark / 4)}": f"{round(self.max_threshold * mark / 4)}"
                        for mark in range(5)
                    },
                    id="threshold_id",
                ),
            ],
            className="filter_dashed",
        )

        self.components["filter"]["max_contrib"] = dbc.Col(
            [
                dbc.Label("Features to display: ", id="max_contrib_label"),
                dcc.Slider(
                    min=1,
                    max=min(self.settings["features"], len(self.explainer.x_init.columns)),
                    step=1,
                    value=min(self.settings["features"], len(self.explainer.x_init.columns)),
                    id="max_contrib_id",
                ),
            ],
            className="filter_dashed",
        )

        self.components["filter"]["positive_contrib"] = dbc.Col(
            [
                dbc.Row(
                    [
                        dbc.Label("Contributions to display:", style={"font-size": "95%"}),
                    ]
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            dbc.Checklist(
                                options=[{"label": "Positive", "value": 1}],
                                value=[1],
                                inline=True,
                                id="check_id_positive",
                                # define the font-size style
                                style={"font-size": "82%"},
                            ),
                            style={"display": "inline-block"},
                        ),
                        dbc.Col(
                            dbc.Checklist(
                                options=[{"label": "Negative", "value": 1}],
                                value=[1],
                                inline=True,
                                id="check_id_negative",
                                # define the font-size style
                                style={"font-size": "82%"},
                            ),
                            style={"display": "inline-block"},
                            align="center",
                        ),
                    ],
                    className="g-0",
                    justify="center",
                ),
            ],
            className="filter_dashed",
        )

        self.components["filter"]["masked_contrib"] = dbc.Col(
            [
                dbc.Label("Feature(s) to mask:"),
                dcc.Dropdown(
                    options=[
                        {"label": key, "value": value}
                        for key, value in sorted(self.explainer.inv_features_dict.items(), key=lambda item: item[0])
                    ],
                    value="",
                    multi=True,
                    searchable=True,
                    id="masked_contrib_id",
                ),
            ],
            className="filter_dashed",
        )

    def make_skeleton(self):
        """
        Describe the app skeleton (bootstrap grid) and initialize components containers
        """
        self.skeleton["navbar"] = dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(
                            html.A(
                                dbc.Row(
                                    [
                                        dbc.Col([html.Img(src=self.logo, height="40px")], className="col-1"),
                                        dbc.Col([html.H4("Shapash Monitor", id="shapash_title")]),
                                    ],
                                    align="center",
                                    style={"color": self.title_menu_color},
                                ),
                                href="https://github.com/MAIF/shapash",
                                target="_blank",
                            ),
                            # Change md=3 to md=2
                            md=2,
                            align="center",
                            width="100%",
                            style={"padding": "auto"},
                        ),
                        dbc.Col(
                            [
                                html.A(
                                    dbc.Row(
                                        [
                                            html.H3(
                                                truncate_str(self.explainer.title_story, maxlen=40),
                                                id="shapash_title_story",
                                                style={"text-align": "center"},
                                            )
                                        ]
                                    ),
                                    href="https://github.com/MAIF/shapash",
                                    target="_blank",
                                )
                            ],
                            # Change md=3 to md=4
                            md=4,
                            align="center",
                            width="100%",
                            style={"padding": "auto"},
                        ),
                        dbc.Col([self.components["menu"]], align="end", md=6, width="100%"),
                    ],
                    style={"padding": "5px 15px", "verticalAlign": "middle", "width": "auto", "justify": "end"},
                )
            ],
            fluid=True,
            style={"height": "100%", "backgroundColor": self.bkg_color},
        )

        self.skeleton["body"] = dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                dbc.Card(
                                    [
                                        html.Div(
                                            # To drow the global_feature_importance graph
                                            self.draw_component("graph", "global_feature_importance"),
                                            id="card_global_feature_importance",
                                            # Position must be absolute to add the explanation button
                                            style={"position": "absolute"},
                                        ),
                                        dcc.Store(id="clickdata-store"),
                                        html.Div(
                                            [
                                                # Create explanation button on feature importance graph
                                                dbc.Button(
                                                    "?", id="open_feature_importance", size="sm", color="warning"
                                                ),
                                                # Create popover for this button
                                                dbc.Popover(
                                                    "Click here to have more \
                                                  information on Feature Importance graph.",
                                                    target="open_feature_importance",
                                                    body=True,
                                                    trigger="hover",
                                                ),
                                                # Create modal associated to this button
                                                dbc.Modal(
                                                    [
                                                        # Modal title
                                                        dbc.ModalHeader(dbc.ModalTitle("Feature importance")),
                                                        dbc.ModalBody(
                                                            [
                                                                html.Div(
                                                                    # Add explanation
                                                                    dcc.Markdown(self.explanations.feature_importance)
                                                                ),
                                                                # Here to add link in the modal
                                                                html.A(
                                                                    "Click here for more details",
                                                                    href="https://github.com/MAIF/shapash/blob/master/tutorial/plots_and_charts/tuto-plot03-features-importance.ipynb",
                                                                    # open new brother tab
                                                                    target="_blank",
                                                                    style={"color": self.color[0]},
                                                                ),
                                                            ]
                                                        ),
                                                        # button to close the modal
                                                        dbc.ModalFooter(
                                                            dbc.Button(
                                                                "Close", id="close_feature_importance", color="warning"
                                                            )
                                                        ),
                                                    ],
                                                    id="modal_feature_importance",
                                                    centered=True,
                                                    size="lg",
                                                ),
                                            ],
                                            # position must be relative
                                            style={"position": "relative", "left": "96%"},
                                        ),
                                    ]
                                )
                            ],
                            md=5,
                            style={"padding": "10px 10px 0px 10px"},
                        ),
                        dbc.Col(
                            [
                                # Tabs that contain 3 children tab (Dataset,
                                # Dataset Filters and True Value Vs Pedicted Values)
                                dbc.Tabs(
                                    [
                                        # Tab which contains the datatable component
                                        dbc.Tab(
                                            # draw datatable component
                                            self.draw_component("table", "dataset"),
                                            # Tab name
                                            label="Dataset",
                                            className="card",
                                            id="card_dataset",
                                            # Style of the tab
                                            style={"cursor": "pointer"},
                                            label_style={"color": "black", "height": "30px", "padding": "0px 5px"},
                                            # Style when the tab is activated
                                            active_tab_class_name="fw-bold fst-italic",
                                            active_label_style={
                                                "border-top": "3px solid",
                                                "border-top-color": self.color[0],
                                            },
                                        ),
                                        # Tab which contains components to filter the dataset
                                        dbc.Tab(
                                            dbc.Card(
                                                dbc.CardBody(
                                                    html.Div(
                                                        # draw the component
                                                        self.draw_filter_table(),
                                                        id="card_filter_dataset",
                                                        # To add scroll in overflow y
                                                        style={"overflow-y": "scroll", "overflow-x": "hidden"},
                                                    )
                                                ),
                                                style={"height": "24.1rem"},
                                            ),
                                            # Tab name
                                            label="Dataset Filters",
                                            # Style of the tab
                                            label_style={"color": "black", "height": "30px", "padding": "0px 5px"},
                                            tab_style={
                                                "border-left": "2px solid #ddd",
                                                "border-right": "2px solid #ddd",
                                            },
                                            # Style when the tab is activated
                                            active_tab_class_name="fw-bold fst-italic",
                                            active_label_style={
                                                "border-top": "3px solid",
                                                "border-top-color": self.color[0],
                                            },
                                        ),
                                        # Tab which contains prediction picking graph
                                        # and its explanation button
                                        dbc.Tab(
                                            dbc.Card(
                                                [
                                                    html.Div(
                                                        # draw prediction picking graph
                                                        self.draw_component("graph", "prediction_picking"),
                                                        id="card_prediction_picking",
                                                        # Position must be absolute to add
                                                        # the explanation button
                                                        style={"position": "absolute"},
                                                    ),
                                                    html.Div(
                                                        [
                                                            # Create explanation button
                                                            dbc.Button(
                                                                "?",
                                                                id="open_prediction_picking",
                                                                size="sm",
                                                                color="warning",
                                                            ),
                                                            # Create popover for this button
                                                            dbc.Popover(
                                                                "Click here to have more \
                                                         information on Prediction Picking graph.",
                                                                target="open_prediction_picking",
                                                                body=True,
                                                                trigger="hover",
                                                            ),
                                                            # Create modal associated to this button
                                                            dbc.Modal(
                                                                [
                                                                    # Modal title
                                                                    dbc.ModalHeader(
                                                                        dbc.ModalTitle(
                                                                            "True values Vs Predicted values"
                                                                        )
                                                                    ),
                                                                    dbc.ModalBody(
                                                                        [
                                                                            html.Div(
                                                                                # explanation
                                                                                dcc.Markdown(
                                                                                    self.explanations.prediction_picking
                                                                                )
                                                                            ),
                                                                            # Here to add link in the modal
                                                                            html.Div(
                                                                                html.Img(
                                                                                    src="https://github.com/MAIF/shapash/blob/master/docs/_static/shapash_select_subset.gif?raw=true"
                                                                                )
                                                                            ),
                                                                            # Here to add a link in the modal
                                                                            html.A(
                                                                                "Click here for more details",
                                                                                href="https://github.com/MAIF/shapash/blob/master/tutorial/plots_and_charts/tuto-plot06-prediction_plot.ipynb",
                                                                                # open new brother tab
                                                                                target="_blank",
                                                                                style={"color": self.color[0]},
                                                                            ),
                                                                        ]
                                                                    ),
                                                                    dbc.ModalFooter(
                                                                        # button to close the modal
                                                                        dbc.Button(
                                                                            "Close",
                                                                            id="close_prediction_picking",
                                                                            color="warning",
                                                                        )
                                                                    ),
                                                                ],
                                                                id="modal_prediction_picking",
                                                                centered=True,
                                                                size="lg",
                                                            )
                                                            # Position must be relative
                                                        ],
                                                        style={"position": "relative", "left": "97%"},
                                                    ),
                                                ]
                                            ),
                                            # Tab name
                                            label="True Values Vs Predicted Values",
                                            # Style of the tab
                                            label_style={"color": "black", "height": "30px", "padding": "0px 5px"},
                                            # Style when the tab is activated
                                            active_tab_class_name="fw-bold fst-italic",
                                            active_label_style={
                                                "border-top": "3px solid",
                                                "border-top-color": self.color[0],
                                            },
                                        ),
                                    ],
                                    id="tabs",
                                )
                            ],
                            md=7,
                            style={"padding": "10px 10px"},
                        ),
                    ],
                    style={"padding": "10px 10px 0px 10px", "height": "100%"},
                    align="center",
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                # Card which contains feature selector graph
                                # and explanation button
                                dbc.Card(
                                    [
                                        html.Div(
                                            # Draw feature selector graph
                                            self.draw_component("graph", "feature_selector"),
                                            id="card_feature_selector",
                                            # Position must be absolute to
                                            # add explanation button
                                            style={"position": "absolute"},
                                        ),
                                        html.Div(
                                            [
                                                # Create explanation button
                                                dbc.Button("?", id="open_feature_selector", size="sm", color="warning"),
                                                # popover of this button
                                                dbc.Popover(
                                                    "Click here to have more \
                                                         information on Feature Selector graph.",
                                                    target="open_feature_selector",
                                                    body=True,
                                                    trigger="hover",
                                                ),
                                                # Modal of this button
                                                dbc.Modal(
                                                    [
                                                        dbc.ModalHeader(dbc.ModalTitle("Feature selector")),
                                                        dbc.ModalBody(
                                                            [
                                                                html.Div(
                                                                    # explanations
                                                                    dcc.Markdown(self.explanations.feature_selector)
                                                                ),
                                                                # Here to add link
                                                                html.A(
                                                                    "Click here for more details",
                                                                    href="https://github.com/MAIF/shapash/blob/master/tutorial/plots_and_charts/tuto-plot02-contribution_plot.ipynb",
                                                                    # open new brother tab
                                                                    target="_blank",
                                                                    style={"color": self.color[0]},
                                                                ),
                                                            ]
                                                        ),
                                                        dbc.ModalFooter(
                                                            # Button to close modal
                                                            dbc.Button(
                                                                "Close", id="close_feature_selector", color="warning"
                                                            )
                                                        ),
                                                    ],
                                                    id="modal_feature_selector",
                                                    centered=True,
                                                    size="lg",
                                                ),
                                            ],
                                            # position must be relative
                                            style={"position": "relative", "left": "96%"},
                                        ),
                                    ]
                                )
                            ],
                            md=5,
                            align="center",
                            style={"padding": "0px 10px"},
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            [
                                                # Card that contains detail feature graph
                                                # and explanation button
                                                dbc.Card(
                                                    [
                                                        html.Div(
                                                            # draw detail_feature graph
                                                            self.draw_component("graph", "detail_feature"),
                                                            id="card_detail_feature",
                                                            style={"position": "absolute"},
                                                        ),
                                                        html.Div(
                                                            [
                                                                # Create explanation button
                                                                dbc.Button(
                                                                    "?",
                                                                    id="open_detail_feature",
                                                                    size="sm",
                                                                    color="warning",
                                                                ),
                                                                # Popover of this button
                                                                dbc.Popover(
                                                                    "Click here to have more \
                                                         information on Detail Feature graph.",
                                                                    target="open_detail_feature",
                                                                    body=True,
                                                                    trigger="hover",
                                                                ),
                                                                # Modal of this button
                                                                dbc.Modal(
                                                                    [
                                                                        dbc.ModalHeader(
                                                                            dbc.ModalTitle("Detail feature")
                                                                        ),
                                                                        dbc.ModalBody(
                                                                            [
                                                                                html.Div(
                                                                                    # explanations
                                                                                    dcc.Markdown(
                                                                                        self.explanations.detail_feature
                                                                                    )
                                                                                ),
                                                                                # Here to add link on the modal
                                                                                html.A(
                                                                                    "Click here for more details",
                                                                                    href="https://github.com/MAIF/shapash/blob/master/tutorial/plot/tuto-plot01-local_plot-and-to_pandas.ipynb",
                                                                                    # open new brother tab
                                                                                    target="_blank",
                                                                                    style={"color": self.color[0]},
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        dbc.ModalFooter(
                                                                            # Button to close the modal
                                                                            dbc.Button(
                                                                                "Close",
                                                                                id="close_detail_feature",
                                                                                color="warning",
                                                                            )
                                                                        ),
                                                                    ],
                                                                    id="modal_detail_feature",
                                                                    centered=True,
                                                                    size="lg",
                                                                ),
                                                            ],
                                                            # Position must be relative
                                                            style={"position": "relative", "left": "96%"},
                                                        ),
                                                    ]
                                                )
                                            ],
                                            md=8,
                                            align="center",
                                        ),
                                        dbc.Col(
                                            [
                                                html.Div(
                                                    self.draw_filter(),
                                                    className="card_filter",
                                                    id="card_filter",
                                                ),
                                            ],
                                            md=4,
                                            align="center",
                                        ),
                                    ],
                                ),
                            ],
                            md=7,
                            align="center",
                            style={"padding": "0px 10px"},
                        ),
                    ],
                    style={"padding": "10px 5px 10px 10px"},
                ),
            ],
            className="mt-12",
            fluid=True,
            # To drop the x scroll-bar
            style={"overflow-x": "hidden"},
        )

    def adjust_menu(self):
        """
        Override menu from explainer object depending on
        classification or regression case.
        """
        on_style = {
            "backgroundColor": self.color[0],
            "color": self.bkg_color,
            "margin-top": "0.5rem",
            "margin-right": "0.5rem",
        }
        off_style = {"display": "none"}
        if self.explainer._case == "classification":
            self.components["menu"]["select_label"].options = [
                {"label": f"{self.explainer.label_dict[label] if self.explainer.label_dict else label}", "value": label}
                for label in self.explainer._classes
            ]
            self.components["menu"]["classification_badge"].style = on_style
            self.components["menu"]["regression_badge"].style = off_style
            self.components["menu"]["select_label"].value = self.label

        elif self.explainer._case == "regression":
            self.components["menu"]["classification_badge"].style = off_style
            self.components["menu"]["regression_badge"].style = on_style
            self.components["menu"]["select_collapse"].is_open = False

        else:
            raise ValueError(f"No rule defined for explainer case : {self.explainer._case}")

    def draw_component(self, component_type, component_id, title=None):
        """
        Method which return a component from a type and id.
        It's the method to insert component inside component container.
        Parameters
        ----------
        component_type : string
            Type of the component. Can be table, graph, ...
        component_id : string
            Id of the component. It must be unique.
        title : string, optional
            by default None
        Returns
        -------
        list
            list of components
            (combining for example Graph + embed button to get fullscreen
             details)
        """
        component = [html.H4(title)] if title else []
        component.append(self.components[component_type][component_id])
        component.append(
            html.A(
                html.I("fullscreen", className="material-icons tiny", style={"marginTop": "8px", "marginLeft": "1px"}),
                id=f"ember_{component_id}",
                className="dock-expand",
                **{"data-component-type": component_type},
                # Get components'id
                **{"data-component-id": component_id},
            )
        )
        return component

    def draw_filter_table(self):
        """
        Method which returns the filter dataset components block.
        Returns
        -------
            component
        """
        return self.components["filter"]["filter_dataset"]

    def draw_filter(self):
        """
        Method which returns filter components block for local
        contributions plot.
        Returns
        -------
        list
            list of components
        """
        filter = [
            dbc.Container(
                [
                    dbc.Row([self.components["filter"]["index"]], align="center", style={"height": "4rem"}),
                    dbc.Row([self.components["filter"]["threshold"]], align="center", style={"height": "5rem"}),
                    dbc.Row([self.components["filter"]["max_contrib"]], align="center", style={"height": "5rem"}),
                    dbc.Row([self.components["filter"]["positive_contrib"]], align="center", style={"height": "4rem"}),
                    dbc.Row([self.components["filter"]["masked_contrib"]], align="center"),
                ],
            ),
        ]
        return filter

    @staticmethod
    def select_point(figure, click_data):
        """
        Method which set the selected point in graph component
        corresponding to click_data.
        """
        if click_data:
            curve_id = click_data["points"][0]["curveNumber"]
            point_id = click_data["points"][0]["pointIndex"]
            for curve in range(len(figure["data"])):
                figure["data"][curve].selectedpoints = [point_id] if curve == curve_id else []

    def callback_fullscreen_buttons(self):
        """
        Initialize callbacks for each fullscreen button
        the callback alter style of the component (height, ...)
        Returns
        -------
        dict
            Style of the component
        """
        app = self.app
        components_to_init = {graph: "graph" for graph in self.components["graph"].keys()}
        components_to_init["dataset"] = "table"
        for component_id, component_type in components_to_init.items():
            component_property = "style" if component_type == "graph" else "style_table"

            @app.callback(
                [
                    Output(f"card_{component_id}", "style"),
                    Output(f"{component_id}", component_property),
                ],
                [
                    Input(f"ember_{component_id}", "n_clicks"),
                    Input(f"ember_{component_id}", "data-component-type"),
                    Input(f"ember_{component_id}", "data-component-id"),
                ],
            )
            def ember(click, data_component_type, data_component_id):
                """
                Function used to set style of cards and components.
                Prediction picking graph style is different than the other
                graph because it is placed in a tab.
                ---------------------------------------------------------------
                click: click on zoom button
                data_component_type: component type
                data_component_id: component id
                --------------------------------------------------------------
                return style of cards and style of components
                """
                click = 2 if click is None else click
                toggle_on = True if click % 2 == 0 else False
                if toggle_on:
                    # Style for graph
                    style_component = {"height": "21.6rem"}
                    this_style_card = {
                        "height": "22rem",
                        "zIndex": 900,
                    }
                    # Style for prediction picking graph
                    if data_component_id == "prediction_picking":
                        style_component = {
                            "height": "20.6rem",
                        }
                        this_style_card = {
                            "height": "20.8rem",
                            "zIndex": 901,
                        }
                    # Style for the Dataset
                    if data_component_type == "table":
                        style_component = {
                            "maxHeight": "23rem",
                        }
                        this_style_card = {
                            "height": "24.1rem",
                            "zIndex": 900,
                        }
                    return this_style_card, style_component

                else:
                    # Style when zoom button is clicked
                    this_style_card = {
                        "height": "70vh",
                        "width": "auto",
                        "zIndex": 998,
                        "position": "fixed",
                        "top": "55px",
                        "bottom": 0,
                        "left": 0,
                        "right": 0,
                    }
                    style_component = {
                        "height": "89vh",
                        "maxHeight": "89vh",
                    }
                    return this_style_card, style_component

    def init_callback_settings(self):
        """Callback settings initialization"""
        app = self.app
        self.components["settings"]["input_rows"]["rows"].value = self.settings["rows"]
        self.components["settings"]["input_points"]["points"].value = self.settings["points"]
        self.components["settings"]["input_features"]["features"].value = self.settings["features"]
        self.components["settings"]["input_violin"]["violin"].value = self.settings["violin"]

        for id in self.settings.keys():

            @app.callback([Output(f"{id}", "valid"), Output(f"{id}", "invalid")], [Input(f"{id}", "value")])
            def update_valid(value):
                """
                actualise valid and invalid icon in input component
                Parameters
                ----------
                value : int
                    value of input component
                Returns
                -------
                    tuple of boolean
                """
                patt = re.compile("^[0-9]*[1-9][0-9]*$")
                if patt.match(str(value)):
                    return True, False
                else:
                    return False, True

        @app.callback(
            Output("modal", "is_open"),
            [Input("settings", "n_clicks"), Input("apply", "n_clicks")],
            [
                State("rows", "valid"),
                State("points", "valid"),
                State("features", "valid"),
                State("violin", "valid"),
            ],
        )
        def toggle_modal(n1, n2, rows, points, features, violin):
            """
            open modal /close modal (only if all input are valid)
            """
            ctx = dash.callback_context
            if ctx.triggered[0]["prop_id"] == "settings.n_clicks":
                if n1 is not None:
                    return True
            else:
                if n2 is not None:
                    if all([rows, points, features, violin]):
                        return False
                    else:
                        return True
            return False

    def callback_generator(self):
        """Generates all the app callbacks"""
        app = self.app

        @app.callback(
            [
                Output("dataset", "data"),
                Output("dataset", "tooltip_data"),
                Output("dataset", "columns"),
                Output("filtered_subset_info", "children"),
                Output("filtered_subset_info", "color"),
            ],
            [
                Input("prediction_picking", "selectedData"),
                Input("modal", "is_open"),
                Input("apply_filter", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
            ],
            [
                State("rows", "value"),
                State("name", "value"),
                State({"type": "var_dropdown", "index": ALL}, "value"),
                State({"type": "var_dropdown", "index": ALL}, "id"),
                State({"type": "dynamic-str", "index": ALL}, "value"),
                State({"type": "dynamic-str", "index": ALL}, "id"),
                State({"type": "dynamic-bool", "index": ALL}, "value"),
                State({"type": "dynamic-bool", "index": ALL}, "id"),
                State({"type": "dynamic-date", "index": ALL}, "start_date"),
                State({"type": "dynamic-date", "index": ALL}, "end_date"),
                State({"type": "dynamic-date", "index": ALL}, "id"),
                State({"type": "lower", "index": ALL}, "value"),
                State({"type": "lower", "index": ALL}, "id"),
                State({"type": "upper", "index": ALL}, "value"),
            ],
        )
        def update_datatable(
            selected_data,
            is_open,
            nclicks_apply,
            nclicks_reset,
            nclicks_del,
            rows,
            name,
            val_feature,
            id_feature,
            val_str_modality,
            id_str_modality,
            val_bool_modality,
            id_bool_modality,
            start_date,
            end_date,
            id_date,
            val_lower_modality,
            id_lower_modality,
            val_upper_modality,
        ):
            """
            This function is used to update the datatable according to sorting,
            filtering and settings modifications.
            ------------------------------------------------------------------
            selected_data: selected data in prediction picking graph
            nclicks_apply: click on Apply Filter button
            nclicks_reset: click on Reset All Filter button
            nclicks_del: click on delete button
            rows: number of rows for subset
            name: name for features name
            val_feature: feature selected to filter
            id_feature: id of feature selected to filter
            val_str_modality: string modalities selected
            id_str_modality: id of string modalities selected
            val_bool_modality: boolean modalities selected
            id_bool_modality: id of boolean modalities selected
            start_date: start dates selected
            end_date: end dates selected
            id_date: id of dates selected
            val_lower_modality: lower values of numeric filter
            id_lower_modality: id of lower modalities of numeric filter
            val_upper_modality: upper values of numeric filter
            ------------------------------------------------------------------
            return
            data: available dataset
            tooltip_data: tooltip of the dataset
            columns: columns of the dataset
            filtered_subset_info: subset size
            filtered_subset_color: subset warning color
            """
            ctx = dash.callback_context
            df = self.round_dataframe
            columns = self.components["table"]["dataset"].columns
            filtered_subset_info = None
            filtered_subset_color = None
            if ctx.triggered[0]["prop_id"] == "modal.is_open":
                if is_open:
                    raise PreventUpdate
                else:
                    self.init_data(rows=rows)
                    if name == [1]:
                        columns = [{"name": i, "id": i} for i in self.special_cols] + [
                            {"name": self.features_dict[i], "id": i}
                            for i in self.dataframe.columns.drop(self.special_cols)
                        ]
                    df = self.round_dataframe
            elif (
                (ctx.triggered[0]["prop_id"] == "prediction_picking.selectedData")
                and (selected_data is not None)
                and (len(selected_data) > 1)
            ):
                df = select_data_from_prediction_picking(self.round_dataframe, selected_data)
            # If click on reset button
            elif ctx.triggered[0]["prop_id"] == "reset_dropdown_button.n_clicks":
                df = self.round_dataframe
            # If click on Apply filter
            elif (
                (ctx.triggered[0]["prop_id"] == "apply_filter.n_clicks")
                | ((ctx.triggered[0]["prop_id"] == "prediction_picking.selectedData") and (selected_data is None))
                | (
                    (ctx.triggered[0]["prop_id"] == "prediction_picking.selectedData")
                    and (
                        selected_data is not None
                        and len(selected_data) == 1
                        and len(selected_data["points"]) != 0
                        and selected_data["points"][0]["curveNumber"] > 0
                    )
                )
            ):
                df = self.round_dataframe.copy()
                feature_id = [id_feature[i]["index"] for i in range(len(id_feature))]
                df = select_data_from_str_filters(
                    df,
                    feature_id,
                    id_str_modality,
                    val_feature,
                    val_str_modality,
                )
                df = select_data_from_bool_filters(
                    df,
                    feature_id,
                    id_bool_modality,
                    val_feature,
                    val_bool_modality,
                )
                df = select_data_from_numeric_filters(
                    df,
                    feature_id,
                    id_lower_modality,
                    val_feature,
                    val_lower_modality,
                    val_upper_modality,
                )
                df = select_data_from_date_filters(
                    df,
                    feature_id,
                    id_date,
                    val_feature,
                    start_date,
                    end_date,
                )
                filtered_subset_info = (
                    f"Subset length: {len(df)} ({int(round(100*len(df)/self.explainer.x_init.shape[0]))}%)"
                )
                if len(df) == 0:
                    filtered_subset_color = "danger"

            elif None not in nclicks_del:
                df = self.round_dataframe
            else:
                raise dash.exceptions.PreventUpdate
            data = df.to_dict("records")
            tooltip_data = [
                {column: {"value": str(value), "type": "text"} for column, value in row.items()}
                for row in df.to_dict("index").values()
            ]
            return (
                data,
                tooltip_data,
                columns,
                filtered_subset_info,
                filtered_subset_color,
            )

        @app.callback(
            [
                Output("global_feature_importance", "figure"),
                Output("global_feature_importance", "clickData"),
                Output("clickdata-store", "data"),
            ],
            [
                Input("select_label", "value"),
                Input("dataset", "data"),
                Input("prediction_picking", "selectedData"),
                Input("apply_filter", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
                Input("card_global_feature_importance", "n_clicks"),
                Input("bool_groups", "on"),
                Input("ember_global_feature_importance", "n_clicks"),
            ],
            [
                State("global_feature_importance", "clickData"),
                State("global_feature_importance", "figure"),
                State("features", "value"),
                State("clickdata-store", "data"),
            ],
        )
        def update_feature_importance(
            label,
            data,
            selected_data,
            apply_filters,
            reset_filter,
            nclicks_del,
            n_clicks,
            bool_group,
            click_zoom,
            clickData,
            figure,
            features,
            clickData_store,
        ):
            """
            update feature importance plot according label, click on graph,
            filters applied and subset selected in prediction picking graph.
            ------------------------------------------------------------
            label: label of data
            data: dataset
            selected_data : data selected on prediction picking graph
            apply_filters: click on apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            n_clicks: click on features importance card
            bool_group: display groups
            click_zoom: click on zoom button
            clickData: click on features importance graph
            figure: figure of Features Importance graph
            features: features value
            clickData_store: previous click on features importance graph
            -------------------------------------------------------------
            return
            figure of Features Importance graph
            click on Features Importance graph
            previous click on Features Importance graph
            """
            ctx = dash.callback_context
            # Zoom is False by Default. It becomes True if we click on it
            zoom_active = get_figure_zoom(click_zoom)
            selection = None
            list_index = self.list_index
            if clickData is not None and (
                ctx.triggered[0]["prop_id"]
                in ["apply_filter.n_clicks", "reset_dropdown_button.n_clicks", "dataset.data"]
                or ("del_dropdown_button" in ctx.triggered[0]["prop_id"] and None not in nclicks_del)
            ):
                clickData = update_click_data_on_subset_changes(clickData)

            selected_feature = (
                self.explainer.inv_features_dict.get(get_feature_from_clicked_data(clickData)) if clickData else None
            )

            if self.explainer.features_groups and bool_group:
                if ctx.triggered[0]["prop_id"] == "card_global_feature_importance.n_clicks":
                    # When we click twice on the same bar this will reset the graph
                    if clickData_store == clickData:
                        selected_feature = None
                    selected_feature = get_feature_from_features_groups(
                        selected_feature, self.explainer.features_groups
                    )
                elif ctx.triggered[0]["prop_id"] == "ember_global_feature_importance.n_clicks":
                    selected_feature = get_feature_from_features_groups(
                        selected_feature, self.explainer.features_groups
                    )
                else:
                    selected_feature = None

            selection = get_indexes_from_datatable(data, list_index)

            group_name = get_group_name(selected_feature, self.explainer.features_groups)

            figure = self.explainer.plot.features_importance(
                max_features=features,
                selection=selection,
                label=label,
                group_name=group_name,
                display_groups=bool_group,
                zoom=zoom_active,
            )
            # Adjust graph with adding x axis title
            MyGraph.adjust_graph_static(figure, x_ax="Mean absolute Contribution")
            figure.layout.clickmode = "event+select"
            if selected_feature:
                self.select_point(figure, clickData)

            # font size can be adapted to screen size
            nb_car = max([len(figure.data[0].y[i]) for i in range(len(figure.data[0].y))])
            figure.update_layout(yaxis=dict(tickfont={"size": min(round(500 / nb_car), 12)}))
            clickData_store = clickData.copy() if clickData is not None else None
            return figure, clickData, clickData_store

        @app.callback(
            Output(component_id="feature_selector", component_property="figure"),
            [
                Input("global_feature_importance", "clickData"),
                Input("dataset", "data"),
                Input("select_label", "value"),
                Input("ember_feature_selector", "n_clicks"),
            ],
            [
                State("points", "value"),
                State("violin", "value"),
                State("global_feature_importance", "figure"),
            ],
        )
        def update_feature_selector(feature, data, label, click_zoom, points, violin, gfi_figure):
            """
            Update feature plot according to label, data,
            selected feature on features importance graph,
            filters and settings modifications
            --------------------------------------------
            feature: click on feature importance graph
            data: dataset
            label: selected label
            click_zoom: click on zoom button
            points: points value in setting
            violin: violin value in setting
            gfi_figure: figure of Features Importance graph
            ---------------------------------------------
            return
            fs_figure: feature selector graph
            """
            # Zoom is False by Default. It becomes True if we click on it
            zoom_active = get_figure_zoom(click_zoom)
            subset = None
            list_index = self.list_index
            if feature is not None:
                selected_feature = get_feature_from_clicked_data(feature)
            else:
                selected_feature = self.selected_feature

            if feature is not None and feature["points"][0]["curveNumber"] == 0 and len(gfi_figure["data"]) == 2:
                subset = get_indexes_from_datatable(data, list_index)
            else:
                subset = None

            fs_figure = self.explainer.plot.contribution_plot(
                col=selected_feature,
                selection=subset,
                label=label,
                violin_maxf=violin,
                max_points=points,
                zoom=zoom_active,
            )

            fs_figure["layout"].clickmode = "event+select"
            # Adjust graph with adding x and y axis titles
            MyGraph.adjust_graph_static(
                fs_figure,
                # x_ax=truncate_str(self.layout.selected_feature, 110),
                x_ax=truncate_str(selected_feature, 110),
                y_ax="Contribution",
            )
            return fs_figure

        @app.callback(
            [Output("index_id", "value"), Output("index_id", "n_submit")],
            [
                Input("feature_selector", "clickData"),
                Input("prediction_picking", "clickData"),
                Input("dataset", "active_cell"),
                Input("apply_filter", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
            ],
            [State("dataset", "data"), State("index_id", "value")],  # Get the current value of the index
        )
        def update_index_id(
            click_data, prediction_picking, cell, apply_filters, reset_filter, nclicks_del, data, current_index_id
        ):
            """
            This function is used to update index value according to
            active cell, filters and click data on feature plot or on
            prediction picking graph.
            ----------------------------------------------------------------
            click_data: click on feature selector
            prediction_picking: click on prediction picking graph
            cell: selected sell on dataset
            apply_filters: click on Apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            data: dataset
            current_index_id: the current value of the index
            ----------------------------------------------------------------
            return
            selected index id
            boolean n_submit
            """
            ctx = dash.callback_context
            selected = None
            if ctx.triggered[0]["prop_id"] == "feature_selector.clickData":
                selected = click_data["points"][0]["customdata"][1]
            elif ctx.triggered[0]["prop_id"] == "prediction_picking.clickData":
                selected = prediction_picking["points"][0]["customdata"]
            elif ctx.triggered[0]["prop_id"] == "dataset.active_cell":
                selected = data[cell["row"]]["_index_"]
            elif ("del_dropdown_button" in ctx.triggered[0]["prop_id"]) & (None in nclicks_del):
                selected = current_index_id
            return selected, True

        @app.callback(Output("threshold_label", "children"), [Input("threshold_id", "value")])
        def update_threshold_label(value):
            """
            update threshold label
            """
            return f"Threshold: {value}"

        @app.callback(Output("max_contrib_label", "children"), [Input("max_contrib_id", "value")])
        def update_max_contrib_label(value):
            """
            update max_contrib label
            """
            return f"Features to display: {value}"

        @app.callback(
            [Output("max_contrib_id", "value"), Output("max_contrib_id", "max"), Output("max_contrib_id", "marks")],
            [Input("modal", "is_open")],
            [State("features", "value"), State("max_contrib_id", "value")],
        )
        def update_max_contrib_id(is_open, features, value):
            """
            update max contrib component layout after settings modifications
            """
            ctx = dash.callback_context
            if ctx.triggered[0]["prop_id"] == "modal.is_open":
                if is_open:
                    raise PreventUpdate
                else:
                    value, max, marks = update_features_to_display(features, len(self.explainer.x_init.columns), value)
                    return value, max, marks

        @app.callback(
            Output(component_id="detail_feature", component_property="figure"),
            [
                Input("threshold_id", "value"),
                Input("max_contrib_id", "value"),
                Input("check_id_positive", "value"),
                Input("check_id_negative", "value"),
                Input("masked_contrib_id", "value"),
                Input("select_label", "value"),
                Input("validation", "n_clicks"),
                Input("bool_groups", "on"),
                Input("ember_detail_feature", "n_clicks"),
            ],
            [State("index_id", "value"), State("dataset", "data")],
        )
        def update_detail_feature(
            threshold,
            max_contrib,
            positive,
            negative,
            masked,
            label,
            validation_click,
            bool_group,
            click_zoom,
            index,
            data,
        ):
            """
            update local explanation plot according to app changes.
            -------------------------------------------------------
            threshold: threshold
            max_contrib: max contribution
            positive: boolean
            negative: boolean
            masked: feature(s) to mask
            label: label
            cell: selected cell
            click_data: click on feature selector graph
            prediction_picking: click on prediction picking graph
            validation_click: click on validation
            bool_group: boolean
            click_zoom: click on zoom button
            index: selected index
            data: the dataset
            --------------------------------------------------------
            return
            detail feature graph
            """
            # Zoom is False by Default. It becomes True if we click on it
            zoom_active = get_figure_zoom(click_zoom)
            selected = index
            if check_row(data, selected) is None:
                selected = None
            threshold = threshold if threshold != 0 else None
            sign = get_feature_contributions_sign_to_show(positive, negative)
            self.explainer.filter(
                threshold=threshold,
                features_to_hide=masked,
                positive=sign,
                max_contrib=max_contrib,
                display_groups=bool_group,
            )
            figure = self.explainer.plot.local_plot(
                index=selected,
                label=label,
                show_masked=True,
                yaxis_max_label=8,
                display_groups=bool_group,
                zoom=zoom_active,
            )
            if selected is not None:
                # Adjust graph with adding x axis titles
                MyGraph.adjust_graph_static(figure, x_ax="Contribution")
                # font size can be adapted to screen size
                list_yaxis = [figure.data[i].y[0] for i in range(len(figure.data))]
                # exclude new line with labels of y axis
                if list_yaxis != []:
                    list_yaxis = [x.split("<br />")[0] for x in list_yaxis]
                    nb_car = max([len(x) for x in list_yaxis])
                    figure.update_layout(yaxis=dict(tickfont={"size": min(round(500 / nb_car), 12)}))
            return figure

        @app.callback(
            Output("validation", "n_clicks"),
            [Input("index_id", "n_submit")],
        )
        def click_validation(n_submit):
            """
            submit index selection
            """
            if n_submit:
                return 1
            else:
                raise PreventUpdate

        @app.callback(
            Output("id_card", "style"),
            Output("id_card_body", "children"),
            Output("id_card_title_contrib", "children"),
            [
                Input("index_id", "n_submit"),
                Input("select_label", "value"),
                Input("select_id_card_sorting", "value"),
                Input("select_id_card_order", "value"),
            ],
            [
                State("dataset", "data"),
                State("index_id", "value"),
            ],
        )
        def update_id_card(n_submit, label, sort_by, order, data, index):
            """
            Update identity card and display button.
            Parameters
            ----------
            n_submit : boolean
            data : the dataset
            label : selected label for classification
            sort_by : identity card column to sort by, data column labels or contribution
            order : order to sort by, ascending or descending
            index : selected index
            Returns
            -------
            style to display button and children body for modal.
            """
            selected = check_row(data, index)
            title_contrib = "Contribution"
            if n_submit and selected is not None:
                selected_row = get_id_card_features(data, selected, self.special_cols, self.features_dict)
                if self.explainer._case == "classification":
                    if label is None:
                        label = -1
                    label_num, _, label_value = self.explainer.check_label_name(label)
                    selected_contrib = get_id_card_contrib(
                        self.explainer.data, index, self.explainer.features_dict, self.explainer.columns_dict, label_num
                    )
                    proba = self.explainer.plot.local_pred(index, label_num)
                    title_contrib = f"Contribution: {label_value} ({proba.round(2):.2f})"
                    _, _, predicted_label_value = self.explainer.check_label_name(
                        selected_row.loc["_predict_", "feature_value"]
                    )
                    selected_row.loc["_predict_", "feature_value"] = predicted_label_value
                else:
                    selected_contrib = get_id_card_contrib(
                        self.explainer.data,
                        index,
                        self.explainer.features_dict,
                        self.explainer.columns_dict,
                    )

                selected_data = create_id_card_data(
                    selected_row,
                    selected_contrib,
                    sort_by,
                    order,
                    self.special_cols,
                    self.explainer.additional_features_dict,
                )

                children = create_id_card_layout(selected_data, self.explainer.additional_features_dict)

                return {"display": "flex", "margin-left": "auto", "margin-right": 0}, children, title_contrib
            else:
                return {"display": "none"}, [], title_contrib

        @app.callback(
            Output("modal_id_card", "is_open"),
            [Input("id_card", "n_clicks"), Input("close_id_card", "n_clicks")],
            [State("modal_id_card", "is_open")],
        )
        def toggle_modal_id_card(n1, n2, is_open):
            """
            Open and close identity card modal.
            Parameters
            ----------
            n1 : click on button to open
            n2 : click on button to close
            is_open : True if open else False
            Returns
            -------
            boolean True if open else False
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            [
                Output("dataset", "style_data_conditional"),
                Output("dataset", "style_filter_conditional"),
                Output("dataset", "style_header_conditional"),
                Output("dataset", "style_cell_conditional"),
            ],
            [Input("validation", "n_clicks")],
            [State("dataset", "data"), State("index_id", "value")],
        )
        def datatable_layout(validation, data, index):
            ctx = dash.callback_context
            if ctx.triggered[0]["prop_id"] == "validation.n_clicks" and validation is not None:
                pass
            else:
                raise PreventUpdate

            style_data_conditional = [
                {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"},
                {
                    "if": {"state": "selected"},
                    "border": f"1px solid {self.color[0]}",
                },
            ]
            style_filter_conditional = []
            style_header_conditional = [{"if": {"column_id": c}, "fontWeight": "bold"} for c in self.special_cols] + [
                {"if": {"column_id": c}, "font-style": "italic"}
                for c in self.dataframe
                if c in self.explainer.additional_features_dict
            ]
            style_cell_conditional = [
                {"if": {"column_id": c}, "width": "70px", "fontWeight": "bold"} for c in self.special_cols
            ]

            selected = check_row(data, index)
            if selected is not None:
                style_data_conditional += [{"if": {"row_index": selected}, "backgroundColor": self.color[0]}]

            return style_data_conditional, style_filter_conditional, style_header_conditional, style_cell_conditional

        @app.callback(
            Output("prediction_picking", "figure"),
            Output("prediction_picking", "selectedData"),
            [
                Input("dataset", "data"),
                Input("apply_filter", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
                Input("select_label", "value"),
                Input("modal", "is_open"),
                Input("ember_prediction_picking", "n_clicks"),
            ],
            [State("points", "value"), State("violin", "value"), State("prediction_picking", "selectedData")],
        )
        def update_prediction_picking(
            data, apply_filters, reset_filter, nclicks_del, label, is_open, click_zoom, points, violin, selectedData
        ):
            """
            Update feature plot according to label, data,
            selected feature and settings modifications
            ------------------------------------------------
            data: the dataset
            apply_filters: click on apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            label: selected label
            is_open: modal
            click_zoom: click on zoom button
            points: number of points
            violin: number of violin plot
            -------------------------------------------------
            return
            prediction picking graph
            """
            ctx = dash.callback_context
            # Filter subset
            subset = None
            if (
                ctx.triggered[0]["prop_id"] == "apply_filter.n_clicks"
                or ctx.triggered[0]["prop_id"] == "reset_dropdown_button.n_clicks"
                or ("del_dropdown_button" in ctx.triggered[0]["prop_id"] and None not in nclicks_del)
            ):
                selectedData = None
            if selectedData is not None and len(selectedData["points"]) > 0:
                raise PreventUpdate
            else:
                subset = get_indexes_from_datatable(data)

            figure = self.explainer.plot.scatter_plot_prediction(selection=subset, max_points=points, label=label)
            if self.explainer.y_target is not None:
                figure["layout"].clickmode = "event+select"
                # Adjust graph with adding x and y axis titles
                MyGraph.adjust_graph_static(figure, x_ax="True Values", y_ax="Predicted Values")

            return figure, selectedData

        @app.callback(
            Output("modal_feature_importance", "is_open"),
            [Input("open_feature_importance", "n_clicks"), Input("close_feature_importance", "n_clicks")],
            [State("modal_feature_importance", "is_open")],
        )
        def toggle_modal_feature_importancet(n1, n2, is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on feature_importance graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_feature_selector", "is_open"),
            [Input("open_feature_selector", "n_clicks"), Input("close_feature_selector", "n_clicks")],
            [State("modal_feature_selector", "is_open")],
        )
        def toggle_modal_feature_selector(n1, n2, is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on feature_selector graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_detail_feature", "is_open"),
            [Input("open_detail_feature", "n_clicks"), Input("close_detail_feature", "n_clicks")],
            [State("modal_detail_feature", "is_open")],
        )
        def toggle_modal_detail_feature(n1, n2, is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on detail_feature graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_prediction_picking", "is_open"),
            [Input("open_prediction_picking", "n_clicks"), Input("close_prediction_picking", "n_clicks")],
            [State("modal_prediction_picking", "is_open")],
        )
        def toggle_modal_prediction_picking(n1, n2, is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on prediction_picking graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_filter", "is_open"),
            [Input("open_filter", "n_clicks"), Input("close_filter", "n_clicks")],
            [State("modal_filter", "is_open")],
        )
        def toggle_modal_filters(n1, n2, is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on Dataset Filters Tab
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        # Add or remove plot blocs in the 'dropdowns_container'
        @app.callback(
            Output("dropdowns_container", "children"),
            [
                Input("add_dropdown_button", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
            ],
            [State("dropdowns_container", "children"), State("name", "value")],
        )
        def layout_filter(n_clicks_add, n_clicks_rm, n_clicks_reset, currents_filters, name):
            """
            Function used to create filter blocs in the dropdowns_container.
            Each bloc will contains:
                -label
                -dropdown button to select feature to filter
                -div which will contains modalities
                -delete button
            ---------------------------------------------------------------
            n_clicks_add: click on add filter
            n_clicks_reset: click on reset filter button
            n_click_del: click on delete button
            children: information on dropdown container
            name: name for feature name
            ---------------------------------------------------------------
            return
                filter blocs
            """
            # Context and init handling (no action)
            ctx = dash.callback_context
            if not ctx.triggered:
                raise dash.exceptions.PreventUpdate
            button_id = ctx.triggered[0]["prop_id"].split(".")[0]

            options = get_feature_filter_options(self.dataframe, self.features_dict, self.special_cols)

            # Creation of a new graph
            if button_id == "add_dropdown_button":
                subset_filter = create_dropdown_feature_filter(n_clicks_add, options)
                return currents_filters + [subset_filter]
            # Removal of all existing filters
            elif button_id == "reset_dropdown_button":
                return [html.Div(id={"type": "bloc_div", "index": 0}, children=[])]
            # Removal of an existing filter
            else:
                filter_id_to_remove = eval(button_id)["index"]
                return [gr for gr in currents_filters if gr["props"]["id"]["index"] != filter_id_to_remove]

        @app.callback(
            Output("reset_dropdown_button", "disabled"),
            [
                Input("add_dropdown_button", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
            ],
        )
        def update_disabled_reset_button(n_click_add, n_click_reset, n_click_del):
            """
            Function used to disabled or not the reset filter button.
            This button is disabled if there is no filter added.
            ---------------------------------------------------------------
            n_click_add: click on add filter button
            n_click_reset: click on reset filter button
            n_click_del: click on delete button
            ---------------------------------------------------------------
            return disabled style
            """
            ctx = dash.callback_context
            if ctx.triggered[0]["prop_id"] == "add_dropdown_button.n_clicks":
                disabled = False
            elif ctx.triggered[0]["prop_id"] == "reset_dropdown_button.n_clicks":
                disabled = True
            elif None not in n_click_del:
                disabled = True
            else:
                disabled = False
            return disabled

        @app.callback(
            Output("apply_filter", "style"),
            Output("filtered_subset_info", "style"),
            [
                Input("add_dropdown_button", "n_clicks"),
                Input("reset_dropdown_button", "n_clicks"),
                Input({"type": "del_dropdown_button", "index": ALL}, "n_clicks"),
            ],
        )
        def update_style_apply_filter_button(n_click_add, n_click_reset, n_click_del):
            """
            Function used to display or not the apply filter button.
            This button is only display if almost one filter was added.
            ---------------------------------------------------------------
            n_click_add: click on add filter button
            n_click_reset: click on reset filter button
            n_click_del: click on delete button
            ---------------------------------------------------------------
            return style of apply filter button
            """
            ctx = dash.callback_context
            if ctx.triggered[0]["prop_id"] == "add_dropdown_button.n_clicks":
                return {}, {"margin-left": "20px"}
            elif ctx.triggered[0]["prop_id"] == "reset_dropdown_button.n_clicks":
                return {"display": "none"}, {"display": "none"}
            elif None not in n_click_del:
                return {"display": "none"}, {"display": "none"}
            else:
                return {}, {"margin-left": "20px"}

        @app.callback(
            Output({"type": "dynamic-output-label", "index": MATCH}, "children"),
            Input({"type": "var_dropdown", "index": MATCH}, "value"),
        )
        def update_label_filter(value):
            """
            Function used to add label to the filters. Label is updated
            when value is not None
            ---------------------------------------------------------------
            value: value selected on the var dropdown button
            ---------------------------------------------------------------
            return label
            """
            if value is not None:
                return html.Label(f"Variable {value} is filtered")
            else:
                return html.Label("Select variable to filter")

        @app.callback(
            Output({"type": "dynamic-output", "index": MATCH}, "children"),
            [
                Input({"type": "var_dropdown", "index": MATCH}, "value"),
                Input({"type": "var_dropdown", "index": MATCH}, "id"),
                Input("add_dropdown_button", "n_clicks"),
            ],
        )
        def display_output(value, id, add_click):
            """
            Function used to create modalities choices. Componenents are different
            according to the type of the selected variable.
            For string variable: component is a dropdown button
            For boolean variable: component is a RadioItems button
            For Integer variable that have less than 20 modalities: component
            is a dropdown button.
            For date variable: component is a DatePickerRange
            Else: components are lower and upper values
            ---------------------------------------------------------------
            value: value selected on the var dropdown button
            id: id of the var dropdown button
            add_click: click on add_dropdown_button
            ---------------------------------------------------------------
            return modalities components. If the component is new, value
            is empty by default.
            """
            # Context and init handling (no action)
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate
            # No update last modalities values if we click on add button
            if ctx.triggered[0]["prop_id"] == "add_dropdown_button.n_clicks":
                raise PreventUpdate
            # Creation on modalities dropdown button
            else:
                if value is not None:
                    new_element = create_filter_modalities_selection(value, id, self.round_dataframe)
                else:
                    new_element = html.Div()
                return new_element



import re

from dash import dcc


class MyGraph(dcc.Graph):
    """Class inherited from dcc.Graph. Add one method for updating graph layout."""

    def __init__(self, figure, id, style=None, **kwds):
        super().__init__(**kwds)
        self.figure = figure
        self.style = style if style is not None else {}
        self.id = id
        # self.config = {'modeBarButtons': {'pan2d': True}}
        if id == "prediction_picking":
            self.config = {
                # Graph is responsive
                "responsive": True,
                "modeBarButtonsToRemove": [
                    "zoomOut2d",
                    "zoomIn2d",
                    "resetScale2d",
                    "hoverClosestCartesian",
                    "hoverCompareCartesian",
                    "toggleSpikelines",
                ],
                "displaylogo": False,
            }
            # 'modeBarStyle': {'orientation': 'v'}, # Deprecated in Dash 1.17.0
        else:
            self.config = {
                # Graph is responsive
                "responsive": True,
                # Graph don't have select box button
                "modeBarButtonsToRemove": [
                    "lasso2d",
                    "zoomOut2d",
                    "zoomIn2d",
                    "resetScale2d",
                    "hoverClosestCartesian",
                    "hoverCompareCartesian",
                    "toggleSpikelines",
                    "select",
                ],
                "displaylogo": False,
            }

    @staticmethod
    def adjust_graph_static(figure, x_ax="", y_ax=""):
        """
        Override graph layout for app use
        ----------------------------------------
        x_ax: title of the x-axis
        y_ax: title of the y-axis
        ---------------------------------------
        """
        new_title = update_title(figure.layout.title.text)
        figure.update_layout(
            autosize=True,
            margin=dict(l=50, r=10, b=10, t=67, pad=0),
            width=None,
            height=None,
            title={
                "y": 0.95,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
                # update title and font-size of the title
                "text": '<span style="font-size: 1.2vw;">' + new_title + "</span>",
            },
        )
        # update x title and font-size of the title
        figure.update_xaxes(title='<span style="font-size: 1vw;">' + x_ax + "</span>", automargin=True)
        # update y title and font-size of the title
        figure.update_yaxes(title='<span style="font-size: 1vw;">' + y_ax + "</span>", automargin=True)


def update_title(title):
    """
    adapt title content the app layout
    Parameters
    ----------
    title : str
        string to ba adapted
    Returns
    -------
    str
    """
    patt = re.compile("^(.+)<span.+?(Predict: .*|Proba: .*)?</span>$")
    try:
        list_non_empty_str_matches = [x for x in patt.findall(title)[0] if x != ""]
        updated = " - ".join(map(str, list_non_empty_str_matches))
    except Exception:
        updated = title
    return updated



import datetime
from typing import Optional, Tuple

import dash_bootstrap_components as dbc
import numpy as np
import pandas as pd
from dash import dcc, html


def select_data_from_prediction_picking(round_dataframe: pd.DataFrame, selected_data: dict) -> pd.DataFrame:
    """Create a subset dataframe from the prediction picking graph selection.

    Parameters
    ----------
    round_dataframe : pd.DataFrame
        Data to sample
    selected_data : dict
        Selected sample in the prediction picking graph

    Returns
    -------
    pd.DataFrame
        Subset dataframe
    """
    row_ids = []
    for p in selected_data["points"]:
        row_ids.append(p["customdata"])
    df = round_dataframe.loc[row_ids]

    return df


def select_data_from_str_filters(
    df: pd.DataFrame,
    feature_id: list,
    id_str_modality: list,
    val_feature: list,
    val_str_modality: list,
) -> pd.DataFrame:
    """Create a subset dataframe from filters.

    Parameters
    ----------
    df : pd.DataFrame
        Data to sample
    feature_id : list
        features ids
    id_str_modality : list
        string features ids
    val_feature : list
        features names
    val_str_modality : list
        string modalities selected

    Returns
    -------
    pd.DataFrame
        Subset dataframe
    """
    # get list of ID
    str_id = [id_str_modality[i]["index"] for i in range(len(id_str_modality))]
    # If there is some filters
    if len(str_id) > 0:
        for i in range(len(feature_id)):
            if feature_id[i] in str_id:
                position = np.where(np.array(str_id) == feature_id[i])[0][0]
                if (position is not None) & (val_str_modality[position] is not None):
                    df = df[df[val_feature[i]].isin(val_str_modality[position])]

    return df


def select_data_from_bool_filters(
    df: pd.DataFrame,
    feature_id: list,
    id_bool_modality: list,
    val_feature: list,
    val_bool_modality: list,
) -> pd.DataFrame:
    """Create a subset dataframe from filters.

    Parameters
    ----------
    df : pd.DataFrame
        Data to sample
    feature_id : list
        features ids
    id_bool_modality : list
        boolean features ids
    val_feature : list
        features names
    val_bool_modality : list
        boolean modalities selected

    Returns
    -------
    pd.DataFrame
        Subset dataframe
    """
    # get list of ID
    bool_id = [id_bool_modality[i]["index"] for i in range(len(id_bool_modality))]
    # If there is some filters
    if len(bool_id) > 0:
        for i in range(len(feature_id)):
            if feature_id[i] in bool_id:
                position = np.where(np.array(bool_id) == feature_id[i])[0][0]
                if (position is not None) & (val_bool_modality[position] is not None):
                    df = df[df[val_feature[i]] == val_bool_modality[position]]

    return df


def select_data_from_numeric_filters(
    df: pd.DataFrame,
    feature_id: list,
    id_lower_modality: list,
    val_feature: list,
    val_lower_modality: list,
    val_upper_modality: list,
) -> pd.DataFrame:
    """Create a subset dataframe from filters.

    Parameters
    ----------
    df : pd.DataFrame
        Data to sample
    feature_id : list
        features ids
    id_lower_modality : list
        numeric features ids
    val_feature : list
        features names
    val_lower_modality : list
        lower values of numeric filter
    val_upper_modality : list
        upper values of numeric filter

    Returns
    -------
    pd.DataFrame
        Subset dataframe
    """
    # get list of ID
    lower_id = [id_lower_modality[i]["index"] for i in range(len(id_lower_modality))]
    # If there is some filters
    if len(lower_id) > 0:
        for i in range(len(feature_id)):
            if feature_id[i] in lower_id:
                position = np.where(np.array(lower_id) == feature_id[i])[0][0]
                if (
                    (position is not None)
                    & (val_lower_modality[position] is not None)
                    & (val_upper_modality[position] is not None)
                ):
                    if val_lower_modality[position] < val_upper_modality[position]:
                        df = df[
                            (df[val_feature[i]] >= val_lower_modality[position])
                            & (df[val_feature[i]] <= val_upper_modality[position])
                        ]

    return df


def select_data_from_date_filters(
    df: pd.DataFrame,
    feature_id: list,
    id_date: list,
    val_feature: list,
    start_date: list,
    end_date: list,
) -> pd.DataFrame:
    """Create a subset dataframe from filters.

    Parameters
    ----------
    round_dataframe : pd.DataFrame
        Data to sample
    id_feature : list
        features ids
    id_date : list
        date features ids
    val_feature : list
        features names
    start_date : list
        start dates selected
    end_date : list
        end dates selected

    Returns
    -------
    pd.DataFrame
        Subset dataframe
    """
    # get list of ID
    date_id = [id_date[i]["index"] for i in range(len(id_date))]
    # If there is some filters
    if len(date_id) > 0:
        for i in range(len(feature_id)):
            if feature_id[i] in date_id:
                position = np.where(np.array(date_id) == feature_id[i])[0][0]
                if (position is not None) & (start_date[position] < end_date[position]):
                    df = df[((df[val_feature[i]] >= start_date[position]) & (df[val_feature[i]] <= end_date[position]))]

    return df


def get_feature_from_clicked_data(click_data: dict) -> str:
    """Get the feature name from the feature importance graph click data.

    Parameters
    ----------
    click_data : dict
        Feature importance graph click data

    Returns
    -------
    str
        Selected feature
    """
    selected_feature = click_data["points"][0]["label"].replace("<b>", "").replace("</b>", "")
    return selected_feature


def get_feature_from_features_groups(selected_feature: Optional[str], features_groups: dict) -> Optional[str]:
    """Get the group feature name of the selected feature.

    Parameters
    ----------
    selected_feature : Optional[str]
        Selected feature
    features_groups : dict
        Groups names and corresponding list of features

    Returns
    -------
    Optional[str]
        Group feature name or selected feature if not in a group.
    """
    list_sub_features = [f for group_features in features_groups.values() for f in group_features]
    if selected_feature in list_sub_features:
        for k, v in features_groups.items():
            if selected_feature in v:
                selected_feature = k
    return selected_feature


def get_group_name(selected_feature: Optional[str], features_groups: Optional[dict]) -> Optional[str]:
    """Get the group feature name if the selected feature is one of the groups.

    Parameters
    ----------
    selected_feature : Optional[str]
        Selected feature
    features_groups : Optional[dict]
        Groups names and corresponding list of features

    Returns
    -------
    Optional[str]
        Group feature name
    """
    group_name = (
        selected_feature if (features_groups is not None and selected_feature in features_groups.keys()) else None
    )
    return group_name


def get_indexes_from_datatable(data: list, list_index: Optional[list] = None) -> Optional[list]:
    """Get the indexes of the data. If list_index is given and is the same length than
    the indexes, there is no subset selected.

    Parameters
    ----------
    data : list
        Data from the table
    list_index : Optional[list], optional
        Default index list to compare the subset with, by default None

    Returns
    -------
    Optional[list]
        Indexes of the data
    """
    indexes = [d["_index_"] for d in data]
    if list_index is not None and (len(indexes) == len(list_index) or len(indexes) == 0):
        indexes = None
    return indexes


def update_click_data_on_subset_changes(click_data: dict) -> dict:
    """Update click data on subset changes to always correspond to the feature selector graph.

    Parameters
    ----------
    click_data : dict
        Feature importance click data

    Returns
    -------
    dict
        Updated feature importance click data
    """
    point = click_data["points"][0]
    point["curveNumber"] = 0
    click_data = {"points": [point]}
    return click_data


def get_figure_zoom(click_zoom: int) -> bool:
    """Get figure zoom from n_clicks

    Parameters
    ----------
    click_zoom : int
        Number of clicks on zoom button

    Returns
    -------
    bool
        zoom active or not
    """
    click = 2 if click_zoom is None else click_zoom
    if click % 2 == 0:
        zoom_active = False
    else:
        zoom_active = True
    return zoom_active


def get_feature_contributions_sign_to_show(positive: list, negative: list) -> Optional[bool]:
    """Get the feature contributions sign to show on plot.

    Parameters
    ----------
    positive : list
        Click on positive contributions
    negative : list
        Click on negative contributions

    Returns
    -------
    Optional[bool]
        Sign to show on plot
    """
    if positive == [1]:
        sign = None if negative == [1] else True
    else:
        sign = False if negative == [1] else None
    return sign


def update_features_to_display(features: int, nb_columns: int, value: int) -> Tuple[int, int, dict]:
    """Update features to display slider.

    Parameters
    ----------
    features : int
        Number of features to plot from the settings
    nb_columns : int
        Number of columns in the data
    value : int
        Number of columns to plot

    Returns
    -------
    Tuple[int, int, dict]
        Number of columns to plot, Number max of columns to plot, Marks in the slider
    """
    max_value = min(features, nb_columns)
    if max_value % 5 == 0:
        nb_marks = min(int(max_value // 5), 10)
    elif max_value % 4 == 0:
        nb_marks = min(int(max_value // 4), 10)
    elif max_value % 3 == 0:
        nb_marks = min(int(max_value // 3), 10)
    elif max_value % 7 == 0:
        nb_marks = min(int(max_value // 7), 10)
    else:
        nb_marks = 2
    marks = {
        f"{round(max_value * feat / nb_marks)}": f"{round(max_value * feat / nb_marks)}"
        for feat in range(1, nb_marks + 1)
    }
    marks["1"] = "1"
    if max_value < value:
        value = max_value

    return value, max_value, marks


def get_id_card_features(data: list, selected: int, special_cols: list, features_dict: dict) -> pd.DataFrame:
    """Get the features of the selected index for the identity card.

    Parameters
    ----------
    data : list
        Data from the table
    selected : int
        Row number of the selected index
    special_cols : list
        Sepcial columns about the index, the prediction...
    features_dict : dict
        Dictionary mapping technical feature names to domain names

    Returns
    -------
    pd.DataFrame
        Dataframe of the features
    """
    selected_row = pd.DataFrame([data[selected]], index=["feature_value"]).T
    selected_row["feature_name"] = selected_row.index.map(lambda x: x if x in special_cols else features_dict[x])
    return selected_row


def get_id_card_contrib(
    data: dict, index: int, features_dict: dict, columns_dict: dict, label_num: int = None
) -> pd.DataFrame:
    """Get the contributions of the selected index for the identity card.

    Parameters
    ----------
    data : dict
        Data from the smart explainer
    index : int
        Index selected
    features_dict : dict
        Dictionary mapping technical feature names to domain names
    columns_dict : dict
        Dictionary mapping integer column number to technical feature names
    label_num : int, optional
        Label num, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe of the contributions
    """
    if label_num is not None:
        contrib = data["contrib_sorted"][label_num].loc[index, :].values
        var_dict = data["var_dict"][label_num].loc[index, :].values
    else:
        contrib = data["contrib_sorted"].loc[index, :].values
        var_dict = data["var_dict"].loc[index, :].values

    var_dict = [features_dict[columns_dict[x]] for x in var_dict]
    selected_contrib = pd.DataFrame([var_dict, contrib], index=["feature_name", "feature_contrib"]).T
    selected_contrib["feature_contrib"] = selected_contrib["feature_contrib"].apply(lambda x: round(x, 4))

    return selected_contrib


def create_id_card_data(
    selected_row: pd.DataFrame,
    selected_contrib: pd.DataFrame,
    sort_by: str,
    order: bool,
    special_cols: list,
    additional_features_dict: dict,
) -> pd.DataFrame:
    """Merge and sort features and contributions dataframes for the identity card.

    Parameters
    ----------
    selected_row : pd.DataFrame
        Dataframe of the features
    selected_contrib : pd.DataFrame
        Dataframe of the contributions
    sort_by : str
        Column to sort by
    order : bool
        Ascending or descending order
    special_cols : list
        Sepcial columns about the index, the prediction...
    additional_features_dict : dict
        Dictionary mapping technical feature names to domain names for additional data

    Returns
    -------
    pd.DataFrame
        Dataframe of the data for the identity card
    """
    selected_data = selected_row.merge(selected_contrib, how="left", on="feature_name")
    selected_data.index = selected_row.index
    selected_data = pd.concat(
        [
            selected_data.loc[special_cols],
            selected_data.drop(index=special_cols + list(additional_features_dict.keys())).sort_values(
                sort_by, ascending=order
            ),
            selected_data.loc[list(additional_features_dict.keys())].sort_values(sort_by, ascending=order),
        ]
    )

    return selected_data


def create_id_card_layout(selected_data: pd.DataFrame, additional_features_dict: dict) -> list:
    """Create the layout of the identity card

    Parameters
    ----------
    selected_data : pd.DataFrame
        Dataframe of the data for the identity card
    additional_features_dict : dict
        Dictionary mapping technical feature names to domain names for additional data

    Returns
    -------
    list
        Layout of the identity card
    """
    children = []
    for _, row in selected_data.iterrows():
        label_style = (
            {"fontWeight": "bold", "font-style": "italic"}
            if row["feature_name"] in additional_features_dict.values()
            else {"fontWeight": "bold"}
        )
        children.append(
            dbc.Row(
                [
                    dbc.Col(dbc.Label(row["feature_name"]), width=3, style=label_style),
                    dbc.Col(dbc.Label(row["feature_value"]), width=5, className="id_card_solid"),
                    dbc.Col(width=1),
                    dbc.Col(
                        dbc.Row(
                            dbc.Label(format(row["feature_contrib"], ".4f"), width="auto", style={"padding-top": 0}),
                            justify="end",
                        ),
                        width=2,
                        className="id_card_solid",
                    )
                    if not np.isnan(row["feature_contrib"])
                    else None,
                ]
            )
        )

    return children


def get_feature_filter_options(dataframe: pd.DataFrame, features_dict: dict, special_cols: list) -> list:
    """Get the columns names options for the filter.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Dataframe
    features_dict : dict
        Dictionary mapping technical feature names to domain names
    special_cols : list
        Sepcial columns about the index, the prediction...

    Returns
    -------
    list
        Options for the filter
    """
    # We use domain name for feature name
    dict_name = [features_dict[i] for i in dataframe.drop(special_cols, axis=1).columns]
    dict_id = [i for i in dataframe.drop(special_cols, axis=1).columns]
    # Create dataframe to sort it by feature_name
    df_feature_name = pd.DataFrame({"feature_name": dict_name, "feature_id": dict_id})
    df_feature_name = df_feature_name.sort_values(by="feature_name").reset_index(drop=True)
    # Options are sorted by feature_name
    options = [{"label": i, "value": i} for i in special_cols] + [
        {"label": df_feature_name.loc[i, "feature_name"], "value": df_feature_name.loc[i, "feature_id"]}
        for i in range(len(df_feature_name))
    ]

    return options


def create_dropdown_feature_filter(n_clicks_add: Optional[int], options: list) -> html.Div:
    """Create a new dropdown for the filter feature selection.

    Parameters
    ----------
    n_clicks_add : Optional[int]
        Number of clicks on the add filter button
    options : list
        Options for the selection

    Returns
    -------
    html.Div
        Div containing the dropdown
    """
    # ID index definition
    if n_clicks_add is None:
        index_id = 0
    else:
        index_id = n_clicks_add
    # Appending a dropdown block to 'dropdowns_container'children
    subset_filter = html.Div(
        id={"type": "bloc_div", "index": index_id},
        children=[
            html.Div(
                [
                    html.Br(),
                    # div which will contains label
                    html.Div(
                        id={"type": "dynamic-output-label", "index": index_id},
                    ),
                ]
            ),
            html.Div(
                [
                    # div with dopdown button to select feature to filter
                    html.Div(
                        dcc.Dropdown(
                            id={"type": "var_dropdown", "index": index_id}, options=options, placeholder="Variable"
                        ),
                        style={"width": "30%"},
                    ),
                    # div which will contains modalities
                    html.Div(id={"type": "dynamic-output", "index": index_id}, style={"width": "50%"}),
                    # Button to delete bloc
                    dbc.Button(
                        id={"type": "del_dropdown_button", "index": index_id}, children="X", color="warning", size="sm"
                    ),
                ],
                style={"display": "flex"},
            ),
        ],
    )

    return subset_filter


def create_filter_modalities_selection(value: str, id: dict, round_dataframe: pd.DataFrame) -> html.Div:
    """Create the modalities filter according to the feature type.

    Parameters
    ----------
    value : str
        feature name
    id : dict
        id of the filter
    round_dataframe : pd.DataFrame
        Dataframe

    Returns
    -------
    html.Div
        Div containing the modalities selection options
    """
    if type(round_dataframe[value].iloc[0]) == bool:
        new_element = html.Div(
            dcc.RadioItems(
                [{"label": val, "value": val} for val in round_dataframe[value].unique()],
                id={"type": "dynamic-bool", "index": id["index"]},
                value=round_dataframe[value].iloc[0],
                inline=False,
            ),
            style={"width": "65%", "margin-left": "20px"},
        )
    elif (type(round_dataframe[value].iloc[0]) == str) | (
        (type(round_dataframe[value].iloc[0]) == np.int64) & (len(round_dataframe[value].unique()) <= 20)
    ):
        new_element = html.Div(
            dcc.Dropdown(
                id={"type": "dynamic-str", "index": id["index"]},
                options=[{"label": i, "value": i} for i in np.sort(round_dataframe[value].unique())],
                multi=True,
            ),
            style={"width": "65%", "margin-left": "20px"},
        )
    elif (type(round_dataframe[value].iloc[0]) is pd.Timestamp) | (
        type(round_dataframe[value].iloc[0]) is datetime.datetime
    ):
        new_element = (
            html.Div(
                dcc.DatePickerRange(
                    id={"type": "dynamic-date", "index": id["index"]},
                    min_date_allowed=round_dataframe[value].min(),
                    max_date_allowed=round_dataframe[value].max(),
                    start_date=round_dataframe[value].min(),
                    end_date=round_dataframe[value].max(),
                ),
                style={"width": "65%", "margin-left": "20px"},
            ),
        )
    else:
        lower_value = 0
        upper_value = 0
        new_element = html.Div(
            [
                dcc.Input(
                    id={"type": "lower", "index": id["index"]},
                    value=lower_value,
                    type="number",
                    style={"width": "60px"},
                ),
                " <= {} in [{}, {}]<= ".format(value, round_dataframe[value].min(), round_dataframe[value].max()),
                dcc.Input(
                    id={"type": "upper", "index": id["index"]},
                    value=upper_value,
                    type="number",
                    style={"width": "60px"},
                ),
            ],
            style={"margin-left": "20px"},
        )

    return new_element



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd

from shapash.utils.check import check_contribution_object, check_model
from shapash.utils.transform import adapt_contributions, get_preprocessing_mapping
from shapash.utils.utils import choose_state


class BaseBackend(ABC):
    """``BaseBackend`` is the base class for all backends.
    All explainability implementations should extend this abstract class
    and implement the methods marked as abstract.
    """

    # class properties
    # --------------------
    # `column_aggregation` defines a way to aggregate local contributions.
    # Default is sum, possible values are 'sum' or 'first'.
    # It allows to compute (column-wise) aggregation of local contributions.
    column_aggregation = "sum"

    # `name` defines the string name of the backend allowing to identify and
    # construct the backend from it.
    name = "base"
    support_groups = True
    supported_cases = ["classification", "regression"]

    def __init__(self, model: Any, preprocessing: Optional[Any] = None):
        """Create a backend instance using a given implementation.

        Parameters
        ----------
        model : any
            Model used.
        preprocessing: category_encoders, ColumnTransformer, list or dict
            The processing apply to the original data.
        """
        self.model = model
        self.preprocessing = preprocessing
        self.explain_data: Any = None
        self.state = None
        self._case, self._classes = check_model(model)
        if self._case not in self.supported_cases:
            raise ValueError(f"Model not supported by the backend as it does not cover {self._case} case")

    @abstractmethod
    def run_explainer(self, x: pd.DataFrame) -> dict:
        """
        Computes local contributions.
        Must be implemented by a child class

        Parameters
        ----------
        x : pd.DataFrame
            The observations dataframe used by the model

        Returns
        -------
        explain_data : dict
            dict containing local contributions
        """
        raise NotImplementedError(
            f"`{self.__class__.__name__}` is a subclass of BaseBackend and "
            f"must implement the `_run_explainer` method"
        )

    def get_local_contributions(
        self, x: pd.DataFrame, explain_data: Any, subset: Optional[List[int]] = None
    ) -> Union[pd.DataFrame, List[pd.DataFrame]]:
        """Get local contributions using the explainer data computed in the `run_explainer`
        method.

        It applies some aggregations and transformations to the result of the `run_explainer` method
        if needed. For example, if there are some one-hot-encoded columns, it automatically checks
        and applies aggregations depending on the result of `_run_explainer` and the `preprocessing`
        (encoder).

        Parameters
        ----------
        x : pd.DataFrame
            The dataframe of observations used by the model.
        explain_data : dict
            The data computed in the `run_explainer` method.
        subset : list
            list of indices on which to get local contributions.

        Returns
        -------
        local_contributions : pd.DataFrame
            The local contributions computed by the backend.
        """
        assert isinstance(explain_data, dict), "The _run_explainer method should return a dict"
        if "contributions" not in explain_data.keys():
            raise ValueError(
                "The _run_explainer method should return a dict"
                " with at least `contributions` key containing "
                "the local contributions"
            )

        local_contributions = explain_data["contributions"]
        if subset is not None:
            local_contributions = local_contributions.loc[subset]
        local_contributions = self.format_and_aggregate_local_contributions(x, local_contributions)
        return local_contributions

    def get_global_features_importance(
        self, contributions: pd.DataFrame, explain_data: Optional[dict] = None, subset: Optional[List[int]] = None
    ) -> Union[pd.Series, List[pd.Series]]:
        """Get global contributions using the explainer data computed in the `run_explainer`
        method.

        Parameters
        ----------
        contributions : pd.DataFrame or list of pd.DataFrame
            The local contributions computed and aggregated by the backend.
        explain_data : dict, optional
            The data computed in the `run_explainer` method.
        subset : list
            list of indices on which to get local contributions.

        Returns
        -------
        pd.Series or list of pd.Series
            The global features importance computed by the backend.
        """
        state = choose_state(contributions)
        if subset is not None:
            if isinstance(contributions, list):
                contributions = [c.loc[subset] for c in contributions]
            else:
                contributions = contributions.loc[subset]
        return state.compute_features_import(contributions)

    def format_and_aggregate_local_contributions(
        self,
        x: pd.DataFrame,
        contributions: Union[pd.DataFrame, np.array, List[pd.DataFrame], List[np.array]],
    ) -> Union[pd.DataFrame, List[pd.DataFrame]]:
        """
        This function allows to format and aggregate contributions in the right format
        (pd.DataFrame or list of pd.DataFrame).

        If a preprocessing exists, it also uses it to change the contributions if needed.

        Parameters
        ----------
        x : pd.DataFrame
            The dataframe of observations used by the model.
        contributions : pd.DataFrame or np.array or list of pd.DataFrame or list of np.array
            Local contributions, or list of local contributions.

        Returns
        -------
        contributions : pd.DataFrame or list of pd.DataFrame
            Contributions formatted and aggregated
        """
        contributions = adapt_contributions(self._case, contributions)
        self.state = choose_state(contributions)
        check_contribution_object(self._case, self._classes, contributions)
        contributions = self.state.validate_contributions(contributions, x)
        contributions_cols = (
            contributions.columns.to_list()
            if isinstance(contributions, pd.DataFrame)
            else contributions[0].columns.to_list()
        )
        if _needs_preprocessing(contributions_cols, x, self.preprocessing):
            contributions = self._apply_preprocessing(contributions)
        return contributions

    def _apply_preprocessing(
        self, contributions: Union[pd.DataFrame, List[pd.DataFrame]]
    ) -> Union[pd.DataFrame, List[pd.DataFrame]]:
        """
        Reconstruct contributions for original features, taken into account a preprocessing.

        Parameters
        ----------
        contributions : object
            Local contributions, or list of local contributions.

        Returns
        -------
        object
            Reconstructed local contributions in the original space. Can be a list.
        """
        if self.preprocessing:
            return self.state.inverse_transform_contributions(
                contributions, self.preprocessing, agg_columns=self.column_aggregation
            )
        else:
            return contributions


def _needs_preprocessing(result_cols, x, preprocessing):
    """
    Checks if preprocessing is needed depending on the preprocessing used.
    """
    mapping = get_preprocessing_mapping(x, preprocessing)
    cols_after_preprocessing = [x for list_c in mapping.values() for x in list_c]
    for col in result_cols:
        if col in cols_after_preprocessing and col not in mapping.keys():
            return True
    return False



"""
Report generation helper module.
"""
import os
from typing import Optional, Union

import pandas as pd
import papermill as pm
from nbconvert import HTMLExporter

from shapash.utils.utils import get_project_root


def execute_report(
    working_dir: str,
    explainer: object,
    project_info_file: str,
    x_train: Optional[pd.DataFrame] = None,
    y_train: Optional[pd.DataFrame] = None,
    y_test: Optional[Union[pd.Series, pd.DataFrame]] = None,
    config: Optional[dict] = None,
    notebook_path: Optional[str] = None,
    kernel_name: Optional[str] = None,
):
    """
    Executes the base_report.ipynb notebook and saves the results in working_dir.

    Parameters
    ----------
    working_dir : str
        Directory in which will be saved the executed notebook.
    explainer : shapash.explainer.smart_explainer.SmartExplainer
        Compiled shapash explainer.
    project_info_file : str
        Path to the file used to display some information about the project in the report.
    x_train : pd.DataFrame
        DataFrame used for training the model.
    y_train : pd.Series or pd.DataFrame
        Series of labels in the training set.
    y_test : pd.Series or pd.DataFrame
        Series of labels in the test set.
    config : dict, optional
        Report configuration options.
    notebook_path : str, optional
        Path to the notebook used to generate the report. If None, the Shapash base report
        notebook will be used.
    kernel_name : str, optional
        Name of the kernel used to generate the report. This parameter can be usefull if
        you have multiple jupyter kernels and that the method does not use the right kernel
        by default.
    """
    if config is None:
        config = {}
    explainer.save(path=os.path.join(working_dir, "smart_explainer.pickle"))
    if x_train is not None:
        x_train.to_csv(os.path.join(working_dir, "x_train.csv"))
    if y_train is not None:
        y_train.to_csv(os.path.join(working_dir, "y_train.csv"))
    if y_test is not None:
        y_test.to_csv(os.path.join(working_dir, "y_test.csv"))
    root_path = get_project_root()
    if notebook_path is None or notebook_path == "":
        notebook_path = os.path.join(root_path, "shapash", "report", "base_report.ipynb")

    pm.execute_notebook(
        notebook_path,
        os.path.join(working_dir, "base_report.ipynb"),
        parameters=dict(dir_path=working_dir, project_info_file=project_info_file, config=config),
        kernel_name=kernel_name,
    )


def export_and_save_report(working_dir: str, output_file: str):
    """
    Exports a previously executed notebook and saves it as a static HTML file.

    Parameters
    ----------
    working_dir : str
        Path to the directory containing the executed notebook.
    output_file : str
        Path to the html file that will be created.
    """

    exporter = HTMLExporter(
        exclude_input=True,
        extra_template_basedirs=[os.path.join(get_project_root(), "shapash", "report", "template")],
        template_name="custom",
        exclude_anchor_links=True,
    )
    (body, resources) = exporter.from_filename(filename=os.path.join(working_dir, "base_report.ipynb"))

    with open(output_file, "w") as file:
        file.write(body)



from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.colors import LinearSegmentedColormap

from shapash.report.common import VarType
from shapash.style.style_utils import get_palette, get_pyplot_color
from shapash.utils.utils import truncate_str


def generate_fig_univariate(
    df_all: pd.DataFrame, col: str, hue: str, type: VarType, colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of any kind of feature
    (continuous, categorical).

    If the feature is categorical and contains too many categories, the smallest
    categories are grouped into a new 'Other' category so that the graph remains
    readable.

    The input dataframe should contain the column of interest and a column that is used
    to distinguish two types of values (ex. 'train' and 'test')

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    type: str
        The type of the series ('continous' or 'categorical')
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    if type == VarType.TYPE_NUM:
        fig = generate_fig_univariate_continuous(df_all, col, hue=hue, colors_dict=colors_dict)
    elif type == VarType.TYPE_CAT:
        fig = generate_fig_univariate_categorical(df_all, col, hue=hue, colors_dict=colors_dict)
    else:
        raise NotImplementedError("Series dtype not supported")
    return fig


def generate_fig_univariate_continuous(
    df_all: pd.DataFrame, col: str, hue: str, colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of a continuous feature.

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette("default")
    g = sns.displot(
        df_all,
        x=col,
        hue=hue,
        kind="kde",
        fill=True,
        common_norm=False,
        palette=get_pyplot_color(colors=colors_dict["report_feature_distribution"]),
    )
    g.set_xticklabels(rotation=30)

    fig = g.fig

    fig.set_figwidth(7)
    fig.set_figheight(4)

    return fig


def generate_fig_univariate_categorical(
    df_all: pd.DataFrame, col: str, hue: str, nb_cat_max: int = 7, colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of a categorical feature.

    If the feature is categorical and contains too many categories, the smallest
    categories are grouped into a new 'Other' category so that the graph remains
    readable.

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    nb_cat_max : int
        The number max of categories to be displayed. If the number of categories
        is greater than nb_cat_max then groups smallest categories into a new
        'Other' category
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette("default")
    df_cat = df_all.groupby([col, hue]).agg({col: "count"}).rename(columns={col: "count"}).reset_index()
    df_cat["Percent"] = df_cat["count"] * 100 / df_cat.groupby(hue)["count"].transform("sum")

    if pd.api.types.is_numeric_dtype(df_cat[col].dtype):
        df_cat = df_cat.sort_values(col, ascending=True)
        df_cat[col] = df_cat[col].astype(str)

    nb_cat = df_cat.groupby([col]).agg({"count": "sum"}).reset_index()[col].nunique()

    if nb_cat > nb_cat_max:
        df_cat = _merge_small_categories(df_cat=df_cat, col=col, hue=hue, nb_cat_max=nb_cat_max)

    fig, ax = plt.subplots(figsize=(7, 4))

    sns.barplot(
        data=df_cat,
        x="Percent",
        y=col,
        hue=hue,
        palette=get_pyplot_color(colors=colors_dict["report_feature_distribution"]),
        ax=ax,
    )

    for p in ax.patches:
        ax.annotate(
            "{:.1f}%".format(np.nan_to_num(p.get_width(), nan=0)),
            xy=(p.get_width(), p.get_y() + p.get_height() / 2),
            xytext=(5, 0),
            textcoords="offset points",
            ha="left",
            va="center",
        )

    # Shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # Put a legend to the right of the current axis
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    # Removes plot borders
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    new_labels = [truncate_str(i.get_text(), maxlen=45) for i in ax.yaxis.get_ticklabels()]
    ax.yaxis.set_ticklabels(new_labels)

    return fig


def _merge_small_categories(df_cat: pd.DataFrame, col: str, hue: str, nb_cat_max: int) -> pd.DataFrame:
    """
    Merges categories of column 'col' of df_cat into 'Other' category so that
    the number of categories is less than nb_cat_max.
    """
    df_cat_sum_hue = df_cat.groupby([col]).agg({"count": "sum"}).reset_index()
    list_cat_to_merge = df_cat_sum_hue.sort_values("count", ascending=False)[col].to_list()[nb_cat_max - 1 :]
    df_cat_other = (
        df_cat.loc[df_cat[col].isin(list_cat_to_merge)].groupby(hue, as_index=False)[["count", "Percent"]].sum()
    )
    df_cat_other[col] = "Other"
    return pd.concat([df_cat.loc[~df_cat[col].isin(list_cat_to_merge)], df_cat_other], axis=0)


def generate_confusion_matrix_plot(
    y_true: Union[np.array, list], y_pred: Union[np.array, list], colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing a confusion matrix that is computed using y_true and
    y_pred parameters.

    Parameters
    ----------
    y_true : array-like
        Ground truth (correct) target values.
    y_pred : array-like
        Estimated targets as returned by a classifier.
    colors_dict : dict
        dict of colors used
    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette("default")
    col_scale = get_pyplot_color(colors=colors_dict["report_confusion_matrix"])
    cmap_gradient = LinearSegmentedColormap.from_list("col_corr", col_scale, N=100)

    df_cm = pd.crosstab(y_true, y_pred, rownames=["Actual"], colnames=["Predicted"])
    fig, ax = plt.subplots(figsize=(7, 4))
    sns.heatmap(df_cm, ax=ax, annot=True, cmap=cmap_gradient, fmt="g")
    return fig



"""
Webapp launch module
This is an example in python how to launch app from explainer
"""
import pandas as pd
from category_encoders import one_hot
from lightgbm import LGBMClassifier, LGBMRegressor
from sklearn.model_selection import train_test_split

from shapash import SmartExplainer

cases = {
    1: "Titanic regression",
    2: "Titanic binary classification",
    3: "Titanic multi class classification",
}

CASE = 2

titanic = pd.read_pickle("tests/data/clean_titanic.pkl")

if CASE == 1:
    features = ["Pclass", "Survived", "Embarked", "Sex"]
    encoder = one_hot.OneHotEncoder(titanic, cols=["Embarked", "Sex"])
    X = titanic[features]
    y = titanic["Age"].to_frame()
    model = LGBMRegressor()

elif CASE == 2:
    features = ["Pclass", "Age", "Embarked", "Sex"]
    encoder = one_hot.OneHotEncoder(titanic, cols=["Embarked", "Sex"])
    X = titanic[features]
    y = titanic["Survived"].to_frame()
    model = LGBMClassifier()

else:
    features = ["Survived", "Age", "Embarked", "Sex"]
    encoder = one_hot.OneHotEncoder(titanic, cols=["Embarked", "Sex"])
    X = titanic[features]
    y = titanic["Pclass"].to_frame()
    model = LGBMClassifier()

titanic_enc = encoder.fit_transform(X)
X_train, X_test, y_train, y_test = train_test_split(
    titanic_enc,
    y,
    test_size=0.2,
)

X_test_ini = X.loc[X_test.index, :]

df = titanic[features + y.columns.to_list()]
df = df.loc[X_test.index, :]
df.reset_index(level=0, inplace=True)

model.fit(X_train, y_train)
y_pred = model.predict(X_test)

y_pred = pd.DataFrame(data=y_pred, columns=y.columns.to_list(), index=X_test.index)

y_target = pd.DataFrame(data=y_test, columns=y.columns.to_list(), index=X_test.index)

response_dict = {0: "Death", 1: "Survival"}
xpl = SmartExplainer(model, preprocessing=encoder, title_story=cases[CASE], label_dict=response_dict)

xpl.compile(X_test, y_pred=y_pred, y_target=y_target)

xpl.init_app()
app = xpl.smartapp.app

if __name__ == "__main__":
    app.run_server(debug=False, host="0.0.0.0", port=8080)



"""
Summarize Module
"""

import warnings

import numpy as np
import pandas as pd
from pandas.core.common import flatten
from sklearn.manifold import TSNE

from shapash.utils.transform import get_features_transform_mapping


def summarize_el(dataframe, mask, prefix):
    """
    Compute a summarized Matrix.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Matrix containing contributions, label or feature names
        that will be summarized
    mask: pd.DataFrame
        Mask to apply during the summary step
    prefix: str
        prefix used for columns name

    Returns
    -------
    pd.DataFrame
        Result of the summarize step
    """
    matrix = dataframe.where(mask.to_numpy()).values.tolist()
    summarized_matrix = [[x for x in ll if str(x) != "nan"] for ll in matrix]
    # Padding to create pd.DataFrame
    max_length = max(len(ll) for ll in summarized_matrix)
    for elem in summarized_matrix:
        elem.extend([np.nan] * (max_length - len(elem)))
    # Create DataFrame
    col_list = [prefix + str(x + 1) for x in list(range(max_length))]
    df_summarized_matrix = pd.DataFrame(summarized_matrix, index=list(dataframe.index), columns=col_list, dtype=object)

    return df_summarized_matrix


def compute_features_import(dataframe):
    """
    Compute a relative features importance, sum of absolute values
     ​​of the contributions for each
     features importance compute in base 100
    Parameters
    ----------
    dataframe: pd.DataFrame
        Matrix containing all contributions

    Returns
    -------
    pd.Series
        feature importance One row by feature,
        index of the serie = dataframe.columns
    """
    feat_imp = dataframe.abs().sum().sort_values(ascending=True)
    tot = feat_imp.sum()
    return feat_imp / tot


def summarize(s_contrib, var_dict, x_sorted, mask, columns_dict, features_dict):
    """
    Compute the summarized contributions of features.

    Parameters
    ----------
    s_contrib: pd.DataFrame
        Matrix containing contributions that will be summarized
    var_dict: pd.DataFrame
        Matrix of feature names that will be summarized
    x_sorted: pd.DataFrame
        Matrix containing the value of each feature
    mask: pd.DataFrame
        Mask to apply during the summary step
    columns_dict:
        Dict of column Names, matches column num with column name
    features_dict:
        Dict of column Label, matches column name with column label

    Returns
    -------
    pd.DataFrame
        Result of the summarize step
    """
    contrib_sum = summarize_el(s_contrib, mask, "contribution_")
    var_dict_sum = summarize_el(var_dict, mask, "feature_").map(
        lambda x: features_dict[columns_dict[x]] if not np.isnan(x) else x
    )
    x_sorted_sum = summarize_el(x_sorted, mask, "value_")

    # Concatenate pd.DataFrame
    summary = pd.concat([contrib_sum, var_dict_sum, x_sorted_sum], axis=1)

    # Ordering columns
    ordered_columns = list(flatten(zip(var_dict_sum.columns, x_sorted_sum.columns, contrib_sum.columns)))
    summary = summary[ordered_columns]
    return summary


def group_contributions(contributions, features_groups):
    """
    Regroup contributions according to features_groups parameter

    Parameters
    ----------
    contributions : pd.DataFrame
        Contributions of each unique feature.
    features_groups : dict
        Python dict that inform which features to regroup.

    Returns
    -------
    contributions : pd.DataFrame
        Contributions with grouped features.
    """
    new_contributions = contributions.copy()
    # Computing features groups that are the sum of their corresponding features contributions
    for group_name in features_groups.keys():
        new_contributions[group_name] = new_contributions[features_groups[group_name]].sum(axis=1)

    # Dropping features that are part of the group of features
    for features_grouped in features_groups.values():
        new_contributions = new_contributions.drop(features_grouped, axis=1)

    return new_contributions


def project_feature_values_1d(feature_values, col, x_init, x_encoded, preprocessing, features_dict, how="tsne"):
    """
    Project feature values of a group of features in 1 dimension.
    If feature_values contains categorical features, use preprocessing to get
    the corresponding encoded variables.

    Parameters
    ----------
    feature_values : pd.DataFrame
        DataFrame that contains the feature values
    col : str
        Name of the group of features.
    preprocessing : category_encoders, ColumnTransformer, list, dict, optional
        Preprocessing used to encode categorical variables.
    x_init : pd.DataFrame
        Pandas dataframe before preprocessing transformations
    x_encoded : pd.DataFrame
        Pandas dataframe after preprocessing transformations
    preprocessing : category_encoders or ColumnTransformer or list or dict or list of dict
        The processing apply to the original data
    features_dict: dict, optional (default: None)
        Dictionary mapping technical feature names to domain names.
    how : str
        Method used to compute groups of features values in one column.

    Returns
    -------
    feature_values : pd.Series
        Series containing the projected feature values.
    """
    # Getting mapping of variables to transform categorical features with corresponding encoded variables
    encoding_mapping = get_features_transform_mapping(x_init, x_encoded, preprocessing)
    col_names_in_xinit = list()
    for c in feature_values.columns:
        col_names_in_xinit.extend(encoding_mapping.get(c, [c]))
    feature_values = x_encoded.loc[feature_values.index, col_names_in_xinit]
    # Project in 1D the feature values
    if how == "tsne":
        try:
            feature_values_proj_1d = TSNE(n_components=1, random_state=1).fit_transform(feature_values)
            feature_values = pd.Series(feature_values_proj_1d[:, 0], name=col, index=feature_values.index)
        except Exception as e:
            warnings.warn(f"Could not project group features values : {e}", UserWarning)
            feature_values = pd.Series(feature_values.iloc[:, 0], name=col, index=feature_values.index)
    elif how == "dict_of_values":
        feature_values.columns = [features_dict.get(x, x) for x in feature_values.columns]
        feature_values = pd.Series(
            feature_values.apply(lambda x: x.to_dict(), axis=1), name=col, index=feature_values.index
        )
    else:
        raise NotImplementedError(f"Unknown method : {how}")
    return feature_values


def compute_corr(df, compute_method):
    """
    Compute correlations between features of given dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame used to compute correlations.
    compute_method : str
        Method used to compute correlations ('phik' or 'pearson').

    Returns
    -------
    pd.DataFrame
    """
    # Remove user warnings (when not enough values to compute correlation).
    warnings.filterwarnings("ignore")
    if compute_method == "phik":
        try:
            from phik import phik_matrix

            return phik_matrix(df, verbose=False)
        except ImportError:
            warnings.warn(
                'Cannot compute phik correlations. Falling back to pearson. Install phik using "pip install phik".',
                UserWarning,
            )
            return df.corr()

    elif compute_method == "pearson":
        return df.corr()
    else:
        raise NotImplementedError(f"Not implemented correlation method : {compute_method}")


def create_grouped_features_values(
    x_init, x_encoded, preprocessing, features_groups, features_dict, how="tsne"
) -> pd.DataFrame:
    """
    Compute projections of groups of features using t-sne.

    Parameters
    ----------
    x_init : pd.DataFrame
        x_encoded dataset with inverse transformation with eventual postprocessing modifications.
    x_encoded : pd.DataFrame
        preprocessed dataset used by the model to perform the prediction.
    preprocessing : category_encoders, ColumnTransformer, list, dict, optional
        Preprocessing used to encode categorical variables.
    features_groups : dict
        Groups names and corresponding list of features
    features_dict: dict, optional (default: None)
        Dictionary mapping technical feature names to domain names.
    how : str
        Method used to compute groups of features values in one column.

    Returns
    -------
    df : pd.DataFrame
        features values with projection used for groups of features
    """
    df = x_init.copy()
    for group in features_groups.keys():
        if not isinstance(features_groups[group], list):
            raise ValueError(f"features_groups[{group}] should be a list of features")
        features_values = x_init[features_groups[group]]
        df[group] = project_feature_values_1d(
            features_values,
            col=group,
            x_init=x_init,
            x_encoded=x_encoded,
            preprocessing=preprocessing,
            features_dict=features_dict,
            how=how,
        )
        for f in features_groups[group]:
            if f in df.columns:
                df.drop(f, axis=1, inplace=True)

    return df



import numpy as np
import pandas as pd
from sklearn.preprocessing import normalize


def _df_to_array(instances):
    """
    Transform inputs into arrays

    Parameters
    ----------
    instances : DataFrame, Series or array
        Input data

    Returns
    -------
    instances : array
        Transformed features
    """
    if isinstance(instances, pd.DataFrame):
        return instances.values
    elif isinstance(instances, pd.Series):
        return np.array([instances.values])
    else:
        return instances


def _compute_distance(x1, x2, mean_vector, epsilon=0.0000001):
    """
    Compute distances between data points by using L1 on normalized data : sum(abs(x1-x2)/(mean_vector+epsilon))

    Parameters
    ----------
    x1 : array
        First vector
    x2 : array
        Second vector
    mean_vector : array
        Each value of this vector is the std.dev for each feature in dataset

    Returns
    -------
    diff : float
        Returns :math:`\\sum(\\frac{|x1-x2|}{mean\\_vector+epsilon})`
    """
    diff = np.sum(np.abs(x1 - x2) / (mean_vector + epsilon))
    return diff


def _compute_similarities(instance, dataset):
    """
    Compute pairwise distances between an instance and all other data points

    Parameters
    ----------
    instance : 1D array
        Reference data point
    dataset : 2D array
        Entire dataset used to identify neighbors

    Returns
    -------
    similarity_distance : array
        V[j] == distance between actual instance and instance j
    """
    mean_vector = np.array(dataset, dtype=np.float32).std(axis=0)
    similarity_distance = np.zeros(dataset.shape[0])

    for j in range(0, dataset.shape[0]):
        # Calculate distance between point and instance j
        dist = _compute_distance(instance, dataset[j], mean_vector)
        similarity_distance[j] = dist

    return similarity_distance


def _get_radius(dataset, n_neighbors, sample_size=500, percentile=95):
    """
    Calculate the maximum allowed distance between points to be considered as neighbors

    Parameters
    ----------
    dataset : DataFrame
        Pool to sample from and calculate a radius
    n_neighbors : int
        Maximum number of neighbors considered per instance
    sample_size : int, optional
        Number of data points to sample from dataset, by default 500
    percentile : int, optional
        Percentile used to calculate the distance threshold, by default 95

    Returns
    -------
    radius : float
        Distance threshold
    """
    # Select 500 points max to sample
    size = min([dataset.shape[0], sample_size])
    # Randomly sample points from dataset
    sampled_instances = dataset[np.random.randint(0, dataset.shape[0], size), :]
    # Define normalization vector
    mean_vector = np.array(dataset, dtype=np.float32).std(axis=0)
    # Initialize the similarity matrix
    similarity_distance = np.zeros((size, size))
    # Calculate pairwise distance between instances
    for i in range(size):
        for j in range(i, size):
            dist = _compute_distance(sampled_instances[i], sampled_instances[j], mean_vector)
            similarity_distance[i, j] = dist
            similarity_distance[j, i] = dist
    # Select top n_neighbors
    ordered_X = np.sort(similarity_distance)[:, 1 : n_neighbors + 1]
    # Select the value of the distance that captures XX% of all distances (percentile)
    return np.percentile(ordered_X.flatten(), percentile)


def find_neighbors(selection, dataset, model, mode, n_neighbors=10):
    """
    For each instance, select neighbors based on 3 criteria:

    1. First pick top N closest neighbors (L1 Norm + st. dev normalization)
    2. Filter neighbors whose model output is too different from instance (see condition below)
    3. Filter neighbors whose distance is too big compared to a certain threshold

    Parameters
    ----------
    selection : list
        Indices of rows to be displayed on the stability plot
    dataset : DataFrame
        Entire dataset used to identify neighbors
    model : model object
        ML model
    mode : str
        "classification" or "regression"
    n_neighbors : int, optional
        Top N neighbors initially allowed, by default 10

    Returns
    -------
    all_neighbors : list of 2D arrays
        Wrap all instances with corresponding neighbors in a list with length (#instances).
        Each array has shape (#neighbors, #features) where #neighbors includes the instance itself.
    """
    instances = dataset.loc[selection].values

    all_neighbors = np.empty((0, instances.shape[1] + 1), float)
    """Filter 1 : Pick top N closest neighbors"""
    for instance in instances:
        c = _compute_similarities(instance, dataset.values)
        # Pick indices of the closest neighbors (and include instance itself)
        neighbors_indices = np.argsort(c)[: n_neighbors + 1]
        # Return instance with its neighbors
        neighbors = dataset.values[neighbors_indices]
        # Add distance column
        neighbors = np.append(neighbors, c[neighbors_indices].reshape(n_neighbors + 1, 1), axis=1)
        all_neighbors = np.append(all_neighbors, neighbors, axis=0)

    # Calculate predictions for all instances and corresponding neighbors
    if mode == "regression":
        # For XGB it is necessary to add columns in df, otherwise columns mismatch
        predictions = model.predict(pd.DataFrame(all_neighbors[:, :-1], columns=dataset.columns))
    elif mode == "classification":
        predictions = model.predict_proba(pd.DataFrame(all_neighbors[:, :-1], columns=dataset.columns))[:, 1]

    # Add prediction column
    all_neighbors = np.append(all_neighbors, predictions.reshape(all_neighbors.shape[0], 1), axis=1)
    # Split back into original chunks (1 chunck = instance + neighbors)
    all_neighbors = np.split(all_neighbors, instances.shape[0])

    """Filter 2 : neighbors with similar blackbox output"""
    # Remove points if prediction is far away from instance prediction
    if mode == "regression":
        # Trick : use enumerate to allow the modifcation directly on the iterator
        for i, neighbors in enumerate(all_neighbors):
            all_neighbors[i] = neighbors[abs(neighbors[:, -1] - neighbors[0, -1]) < 0.1 * abs(neighbors[0, -1])]
    elif mode == "classification":
        for i, neighbors in enumerate(all_neighbors):
            all_neighbors[i] = neighbors[abs(neighbors[:, -1] - neighbors[0, -1]) < 0.1]

    """Filter 3 : neighbors below a distance threshold"""
    # Remove points if distance is bigger than radius
    radius = _get_radius(dataset.values, n_neighbors)

    for i, neighbors in enumerate(all_neighbors):
        # -2 indicates the distance column
        all_neighbors[i] = neighbors[neighbors[:, -2] < radius]
    return all_neighbors


def shap_neighbors(instance, x_encoded, contributions, mode):
    """
    For an instance and corresponding neighbors, calculate various
    metrics (described below) that are useful to evaluate local stability

    Parameters
    ----------
    instance : 2D array
        Instance + neighbours with corresponding features
    x_encoded : DataFrame
        Entire dataset used to identify neighbors
    contributions : DataFrame
        Calculated contribution values for the dataset

    Returns
    -------
    norm_shap_values : array
        Normalized SHAP values (with corresponding sign) of instance and its neighbors
    average_diff : array
        Variability (stddev / mean) of normalized SHAP values (using L1) across neighbors for each feature
    norm_abs_shap_values[0, :] : array
        Normalized absolute SHAP value of the instance
    """
    # Extract SHAP values for instance and neighbors
    # :-2 indicates that two columns are disregarded : distance to instance and model output
    ind = (
        pd.merge(x_encoded.reset_index(), pd.DataFrame(instance[:, :-2], columns=x_encoded.columns), how="inner")
        .set_index(x_encoded.index.name if x_encoded.index.name is not None else "index")
        .index
    )
    # If classification, select contrbutions of one class only
    if mode == "classification" and len(contributions) == 2:
        contributions = contributions[1]
    shap_values = contributions.loc[ind]
    # For neighbors comparison, the sign of SHAP values is taken into account
    norm_shap_values = normalize(shap_values, axis=1, norm="l1")
    # But not for the average impact of the features across the dataset
    norm_abs_shap_values = normalize(np.abs(shap_values), axis=1, norm="l1")
    # Compute the average difference between the instance and its neighbors
    # And replace NaN with 0
    average_diff = np.divide(
        norm_shap_values.std(axis=0),
        norm_abs_shap_values.mean(axis=0),
        out=np.zeros(norm_abs_shap_values.shape[1]),
        where=norm_abs_shap_values.mean(axis=0) != 0,
    )

    return norm_shap_values, average_diff, norm_abs_shap_values[0, :]


def get_min_nb_features(selection, contributions, mode, distance):
    """
    Determine the minimum number of features needed for the prediction \
    of the interpretability method to be *close enough* \
    to the one obtained with all features.

    The closeness is defined via the following distances:

    * For regression:

        .. math::

            distance = \\frac{|output_{allFeatures} - output_{currentFeatures}|}{|output_{allFeatures}|}

    * For classification:

        .. math::

            distance = |output_{allFeatures} - output_{currentFeatures}|

    Parameters
    ----------
    selection : list
        Indices of rows to be displayed on the stability plot
    contributions : DataFrame
        Calculated contribution values for the dataset
    mode : str
        "classification" or "regression"
    distance : float, optional
        How close we want to be from model with all features, by default 0.1 (10%)

    Returns
    -------
    features_needed : list
        List of minimum number of required features (for each instance) to be close enough to the prediction (ex: [4, 7, 8...])
    """
    assert 0 <= distance <= 1

    if mode == "classification" and len(contributions) == 2:
        contributions = contributions[1]
    contributions = contributions.loc[selection].values
    features_needed = []
    # For each instance, add features one by one (ordered by SHAP) until we get close enough
    for i in range(contributions.shape[0]):
        ids = np.flip(np.argsort(np.abs(contributions[i, :])))
        output_value = np.sum(contributions[i, :])

        score = 0
        for j, idx in enumerate(ids):  # noqa: B007
            # j : number of features needed
            # idx : positions of the j top shap values
            score += contributions[i, idx]
            # CLOSE_ENOUGH
            if mode == "regression":
                if abs(score - output_value) < distance * abs(output_value):
                    break
            elif mode == "classification":
                if abs(score - output_value) < distance:
                    break
        features_needed.append(j + 1)
    return features_needed


def get_distance(selection, contributions, mode, nb_features):
    """
    Determine how close we get to the output with all features by using only a subset of them

    Parameters
    ----------
    selection : list
        Indices of rows to be displayed on the stability plot
    contributions : DataFrame
        Calculated contribution values for the dataset
    mode : str
        "classification" or "regression"
    nb_features : int, optional
        Number of features used, by default 5

    Returns
    -------
    distance : array
        List of distances for each instance by using top selected features (ex: np.array([0.12, 0.16...])).

        * For regression:

            * normalized distance between the output of current model and output of full model

        * For classifciation:

            * distance between probability outputs (absolute value)
    """
    if mode == "classification" and len(contributions) == 2:
        contributions = contributions[1]
    assert nb_features <= contributions.shape[1]

    contributions = contributions.loc[selection].values
    top_features = np.array([sorted(row, key=abs, reverse=True) for row in contributions])[:, :nb_features]
    output_top_features = np.sum(top_features[:, :], axis=1)
    output_all_features = np.sum(contributions[:, :], axis=1)

    if mode == "regression":
        distance = abs(output_top_features - output_all_features) / abs(output_all_features)
    elif mode == "classification":
        distance = abs(output_top_features - output_all_features)
    return distance
